"""
Filename: complex.py
Author: William Bowley
Version: 0.2
Clear: X

Description:
    Defines the Complex Packet Class which is
    comprised of a Value, Unit and prefixScale.
"""

from __future__ import annotations

from math import log10, ceil, degrees, trunc, floor
from cmath import phase
from typing import Any
from dataclasses import dataclass
from numpy import complexfloating

from picounits.core.unit import Unit
from picounits.core.scales import PrefixScale

from picounits.core.quantities.packet import Packet
from picounits.core.quantities.scalars.scalar import ScalarPacket

from picounits.lazy_imports import import_factory
from picounits.configuration.picounits import STANDARD_DISPLAY

@dataclass(slots=True, repr=False, unsafe_hash=True)
class ComplexPacket(ScalarPacket):
    """
    A Complex Packet: A prefix, value (Real + Imaginary) and Unit

    NOTE: Prefix is init-only, value is held in absolute form
    """
    def __post_init__(self, prefix: PrefixScale) -> None:
        """ Validates value and unit, then mutates value to Base """
        if not isinstance(self.value, (complex, complexfloating)):
            factory = import_factory("ComplexPacket.__post_init__")

            # Attempts to pass the value to the correct type
            return factory.create(self.value, self.unit, prefix)

        if not isinstance(self.unit, Unit):
            msg = f"Unit must be of type 'Unit', not {type(self.unit)}"
            raise TypeError(msg)

        if not isinstance(prefix, PrefixScale):
            msg = f"Prefix must be of type PrefixScale, not {type(prefix)}"
            raise TypeError(msg)

        # Mutates prefix to PrefixScale.BASE and scales value
        if prefix == PrefixScale.BASE:
            return

        # Ex. Mega (6) - BASE (0) = 6 Hence scaling of 10^6
        prefix_difference = prefix.value - PrefixScale.BASE.value
        factor = 10 ** prefix_difference
        self.value *= factor

    @property
    def name(self) -> str:
        """ Returns the packet name as value + prefix(unit) """
        value, prefix = self._normalize()
        return f"{round(value, STANDARD_DISPLAY)} {prefix}({self.unit.name})"

    @property
    def magnitude(self) -> int | float:
        """ Returns the mathematical absolute value """
        factory = import_factory("ComplexPacket.magnitude")
        return factory.create(abs(self.value), self.unit)

    @property
    def real(self) -> Packet:
        """ Returns the real part of self.value """
        factory = import_factory("ComplexPacket.real")
        return factory.create(self.value.real, self.unit)

    @property
    def imag(self) -> Packet:
        """ Returns the imaginary part of self.value """
        factory = import_factory("ComplexPacket.imag")
        return factory.create(self.value.imag, self.unit)

    def conjugate(self) -> ComplexPacket:
        """ Returns the complex conjugate of self.value """
        value: complex = self.value
        return ComplexPacket(value.conjugate(), self.unit)

    def degree_phase(self) -> Packet:
        """ Returns the phase of self.value in degrees """
        phasor = degrees(phase(self.value))

        factory = import_factory("ComplexPacket.degree_phase")
        return factory.create(phasor, Unit.dimensionless())

    def radians_phase(self) -> Packet:
        """ Returns the phase of self.value in radians """
        phasor = phase(self.value)

        factory = import_factory("ComplexPacket.radians_phase")
        return factory.create(phasor, Unit.dimensionless())

    def degrees_polar(self) -> tuple[Packet, Packet]:
        """ Returns the polar representation of the self.value in degrees """
        phasor = degrees(phase(self.value))
        magnitude = self.magnitude

        factory = import_factory("ComplexPacket.degrees_polar")
        return (
            factory.create(phasor, Unit.dimensionless()), 
            factory.create(magnitude, self.unit)
        )

    def radians_polar(self) -> tuple[Packet, Packet]:
        """ Returns the polar representation of the self.value in radians """
        phasor = phase(self.value)
        magnitude = self.magnitude

        factory = import_factory("ComplexPacket.radians_polar")
        return (
            factory.create(phasor, Unit.dimensionless()),
            factory.create(magnitude, self.unit)
        )

    def _normalize(self) -> tuple[complex, PrefixScale]:
        """ Normalizes the value for packet name representation """
        value = self.value
        if value == 0:
            # Handles division by zero or dimensionless values
            return self.value, PrefixScale.BASE

        # Extraction of exponent
        magnitude = abs(value)
        prefix_power = int(log10(magnitude))

        # Snaps prefix power to multiple of 3
        prefix_power = 3 * (prefix_power // 3)

        # O(n) prefix lookup & calculation of new value
        closest = PrefixScale.from_value(prefix_power)
        value /= 10 ** closest.value

        return value, closest

    def __format__(self, format_spec: str) -> str:
        """ Formats the string based on user input through 'format_spec'"""
        value, prefix = self._normalize()
        formatted_value = format(value, format_spec)

        return f"{formatted_value} {prefix}({self.unit.name})"

    def __ceil__(self) -> Packet:
        """ Defines the behavior for ceiling method """
        real_ceil = ceil(self.value.real)
        imag_ceil = ceil(self.value.imag)

        ceiling = complex(real_ceil, imag_ceil)

        factory = import_factory("ComplexPacket.__ceil__")
        return factory.create(ceiling, self.unit)

    def __floor__(self) -> Packet:
        """ Defines the behavior for floor method """
        new_complex = complex(
            floor(self.value.real),
            floor(self.value.imag)
        )

        factory = import_factory("ComplexPacket.__floor__")
        return factory.create(new_complex, self.unit)


    def __trunc__(self) -> Packet:
        """ Defines the behavior for trunc method """
        new_complex = complex(
            trunc(self.value.real),
            trunc(self.value.imag)
        )

        factory = import_factory("ComplexPacket.__trunc__")
        return factory.create(new_complex, self.unit)


    def __round__(self, ndigits=None):
        """ Defines the behavior for the round operation """
        new_complex = complex(
            round(self.value.real, ndigits),
            round(self.value.imag, ndigits)
        )

        factory = import_factory("ComplexPacket.__round__")
        return factory.create(new_complex, self.unit)

    def __eq__(self, other: Any) -> bool:
        """ Defines the behavior for equality comparison """
        q2 = self._get_other_packet(other)
        if self.unit != q2.unit:
            # Unit equality matters
            return False

        return self.value == q2.value

    def _raise_ordering_error(self) -> None:
        """ Helper for ordering comparison errors """
        msg = (
            "Cannot order complex quantities. "
            "Complex numbers have no natural ordering. "
            "Use abs() to compare magnitudes."
        )
        raise TypeError(msg)

    def __lt__(self, other: Any) -> bool:
        """ Defines the behavior for less than comparison """
        _ = other
        self._raise_ordering_error()

    def __le__(self, other: Any) -> bool:
        """ Defines the behavior for less than or equal to comparison """
        _ = other
        self._raise_ordering_error()

    def __gt__(self, other: Any) -> bool:
        """ Defines the behavior for greater than comparison """
        _ = other
        self._raise_ordering_error()

    def __ge__(self, other: Any) -> bool:
        """ Defines the behavior for greater than or equal to comparison """
        _ = other
        self._raise_ordering_error()

    def __repr__(self) -> str:
        """ Displays the packet name """
        return str(self.name)


"""
Filename: matplot.py
Author: William Bowley
Version: 0.1

Description:
    Graphing script for the analytical lumped
    parameter coil-gun simulation using picounits
"""

from typing import Sequence
import matplotlib.pyplot as plt


def plot(
    time_series: Sequence[float],
    current_series: Sequence[float],
    force_series: Sequence[float],
    velocity_series: Sequence[float]
) -> None:
    """ Plots the coil-gun simulation results """
    plt.style.use("dark_background")
    fig, axes = plt.subplots(3, 1, figsize=(12, 8), sharex=True)

    # Set window title
    fig.canvas.manager.set_window_title("Coilgun Simulation Results")

    # Current vs Time
    axes[0].plot(time_series, current_series, color="#00BFFF", linewidth=2)
    axes[0].set_ylabel("Current (A)", color="white")
    axes[0].set_title("Line Current vs Time", color="white")
    axes[0].tick_params(colors="white")

    # Force vs Time
    axes[1].plot(time_series, force_series, color="#FF4500", linewidth=2)
    axes[1].set_ylabel("Force (N)", color="white")
    axes[1].set_title("Projectile Force vs Time", color="white")
    axes[1].tick_params(colors="white")

    # Velocity vs Time
    axes[2].plot(time_series, velocity_series, color="#7CFC00", linewidth=2)
    axes[2].set_xlabel("Time (s)", color="white")
    axes[2].set_ylabel("Velocity (m/s)", color="white")
    axes[2].set_title("Projectile Velocity vs Time", color="white")
    axes[2].tick_params(colors="white")

    fig.tight_layout(pad=2.0)
    plt.show()



"""
Filename: main.py
Author: William Bowley

Description:
    Example of using picounits to import a material 
    library with units. 
    
    This ensure correctness when parsing values into
    dimensional pipelines such as FEM or lumped parameter models.
"""

from pathlib import Path

from picounits import Quantity as Q, unit_validator, PRESSURE
from picounits.extensions.parser import Parser

# File paths
BASE_DIR = Path(__file__).parent.parent.parent
library = BASE_DIR / "examples/library/materials.uiv"
library_units = BASE_DIR / "examples/library/si_metric.ut"

materials = Parser.open(library, library_units)

# Shows library ontology
materials.tree("materials.uiv")
print("=== Derived parameters from library ===")


# Calculate mixture density of Iron/Copper mix 70%/30%
iron_density = materials.pure_iron.physical.density
copper_density = materials.pure_copper.physical.density

density = iron_density * 7/10 + copper_density * 3 / 10
print(f"70 % Iron 30% Copper density: {density!r}")


# Extract NdFeB grades for usage
@unit_validator(PRESSURE)
def _magnet_energy(br_hc: list[Q]) -> Q:
    """ Calculates the energy density of the magnet """
    remanence, coercivity = br_hc
    return 1/2 * remanence * coercivity

NdFeB = materials.NdFeB
n33, n35, = NdFeB.grades.N33, NdFeB.grades.N35
n33_d, n35_d = _magnet_energy(n33), _magnet_energy(n35)

print(f"Energy Density | N33: {n33_d:.3f}; N35: {n35_d:.3f}")


# Calculates the electrical resistivity of copper
resistivity = 1/materials.pure_copper.electrical.conductivity
print(f"Copper resistivity: {resistivity:.3f}")

print("=======================================")



"""
Filename: main.py
Author: William Bowley

Description:
    Transient thermal solver for the 2D heat equation PDE
    using picounits to show unit-boundary abstractions.
    
    NOTE:
    This is a demonstration model and is not intended to 
    match a refined finite element solver
    
    NOTE:
    This example uses these dependencies:
    To install run the command: pip install matplotlib shapely
"""

from typing import Any
from pathlib import Path

import matplotlib.pyplot as plt
import matplotlib.animation as animation

from engine import PreProcessor, Solver

from picounits import LENGTH as m, TIME as s
from picounits.extensions.parser import Parser


# File paths
BASE_DIR = Path(__file__).parent.parent.parent
library = BASE_DIR / "examples/solver/parameters.uiv"

params = Parser.open(library)

# Build simulation domain and heat source geometry
s_x, s_y = params.source.origin
domain = PreProcessor.rectangle((0*m, 0*m), params.problem.length, params.problem.height)
source = PreProcessor.rectangle((s_x, s_y), params.source.length, params.source.height)

# Meshes the geometry
preprocessor = PreProcessor(domain, source)
problem_solver = Solver(preprocessor)
problem_solver.build(params)

fig, ax = plt.subplots(figsize=(6, 6))

# Initial frame solve
solution_now = problem_solver.solve()

# Initial plot
nodes, triangles = problem_solver.unique_nodes, problem_solver.connectivity
tpc = ax.tripcolor(
    nodes[:, 0], nodes[:, 1], triangles, solution_now,
    shading='gouraud', cmap='RdYlBu_r', vmin=273.15, vmax=500
)
fig.colorbar(tpc, label='Temperature (K)')

ax.set_xlabel(f"Length ({params.problem.length.unit})")
ax.set_ylabel(f"Height ({params.problem.height.unit})")
ax.set_aspect('equal')
ax.set_title("Transient Heat Diffusion")

# Draws the source to the plot
xs, ys = problem_solver.source.exterior.xy
ax.plot(xs, ys, 'k', lw=1.5, label='Heat Source')
ax.legend(loc='upper right')

current_time_val = 0 * s
ax.set_title(f"Transient Heat Diffusion - Time : {current_time_val:.3f}")

def update(frame: Any):
    """ Updates the Transient solution over time"""
    global solution_now, current_time_val
    _ = frame

    solution_now = problem_solver.solve(solution_now)
    current_time_val += params.problem.time_step
    ax.set_title(f"Thermal Diffusion | Time: {current_time_val:.3f}")

    tpc.set_array(solution_now)
    return tpc, ax.title

ani = animation.FuncAnimation(
    fig, update, frames=100, interval=params.problem.time_step.value, blit=False
)
plt.show()


"""
Filename: dimensions.py
Author: William Bowley
Version: 0.4
Clear: N

Description:
    Defines the Dimension dataclass,
    which is comprised of a fundamental unit
    and exponent.

    The fundamental unit 'FBASE' is
    encoded as a enum to enable dimensional
    analysis, custom ordering and symbols
"""

from __future__ import annotations

from enum import Enum, auto
from functools import lru_cache
from dataclasses import dataclass, field
from fractions import Fraction


try:
    # Ensures picounits works even with preferences issues
    from picounits.configuration.config import get_base_symbols, get_base_order
except ImportError:
    def get_base_symbols() -> dict[str, str]:
        """ placeholder """
        return {}

    def get_base_order() -> dict[str, int]:
        """ placeholder """
        return {}


@lru_cache(maxsize=1)
def _symbols() -> dict[str, str]:
    """ Caches the symbols for FBase """
    return get_base_symbols()


@lru_cache(maxsize=1)
def _order() -> dict[str, int]:
    """ Caches the order for FBase """
    return get_base_order()


class FBase(Enum):
    """
    Fundamental SI base dimensions:

    Represents the seven SI base units plus dimensionless quantities.
    Supports custom symbols and ordering via configurations
    """
    TIME    = auto()
    LENGTH  = auto()
    MASS    = auto()
    CURRENT = auto()
    TEMPERATURE = auto()
    AMOUNT  = auto()
    LUMINOSITY  = auto()
    DIMENSIONLESS = auto()

    @property
    def symbol(self) -> str:
        """ Returns the base unit symbol. """
        return _symbols().get(self.name, _SIBASE_SYMBOLS[self])

    @property
    def order(self) -> int:
        """ Returns the base units under consistent order for notation """
        return _order().get(self.name, _ORDER[self])

    @classmethod
    def all_symbols(cls) -> list[str]:
        """ Returns the symbols of all base units """
        return list(_symbols().items())

    @classmethod
    def from_symbol(cls, reference: str) -> FBase | None:
        """ Compares reference symbol with symbol lookup, if not found returns none """
        if not isinstance(reference, str):
            return None

        # Primary: Tries preferred symbols
        for member in cls:
            preferred = _symbols().get(member.name)
            # units symbols are case sensitive
            if preferred and preferred == reference:
                return member

        # Fallback: check standard symbols
        for enum_member, symbol in _SIBASE_SYMBOLS.items():
            # units symbols are case sensitive
            if symbol == reference:
                return enum_member

        return None

    def __str__(self) -> str:
        """ Returns name for __str__ dunder method. """
        return self.name

    def __repr__(self) -> str:
        """ Displays the fundamental base. """
        return f"<FBase type={self.name}, symbol={self.symbol}>"


@dataclass(frozen=True, slots=True)
class Dimension:
    """
    A Physical dimension: Base unit raised to an signed integer or float power.

    Args:
        base: The fundamental unit type (FBase Enum)
        exponent: Integer or float power
    """
    base: FBase = FBase.DIMENSIONLESS
    exponent: int | float = 1

    # Note _name_cache is an implementation detail, zero impact on correctness
    _name_cache: str = field(init=False, repr=False, default="")

    def __post_init__(self) -> None:
        """ Validates base, exponent and may mutate values. """
        if not isinstance(self.base, FBase):
            msg = f"Dimension base must be FBase, not {type(self.base)}"
            raise TypeError(msg)

        if not isinstance(self.exponent, (int, float)):
            msg = (
                "Dimension exponent must be int or float, "
                f"not {type(self.exponent)}"
            )
            raise TypeError(msg)

        # Handle zero exponent: x^0 = 1 (dimensionless)
        if self.exponent == 0 or self.base == FBase.DIMENSIONLESS:
            # NOTE: Mutates only during init to ensure correctness
            object.__setattr__(self, "base", FBase.DIMENSIONLESS)
            object.__setattr__(self, "exponent", 1)

        name = (
            self.base.symbol if self.exponent == 1
            else self.base.symbol + self.superscript
        )

        # NOTE: Caches name to improve performance via base mutation
        object.__setattr__(self, "_name_cache", name)

    @property
    def superscript(self) -> str:
        """ Returns the unicode superscript. """
        if not isinstance(self.exponent, float):
            # Directly translate integers exponent to superscript
            return str(self.exponent).translate(SUPERSCRIPT_MAP)

        # Converts float to fraction
        frac = Fraction(self.exponent).limit_denominator()
        numerator, denominator = frac.numerator, frac.denominator

        # Handles case of natural numbers encoded as floats (e.g 2.0 -> 2)
        if denominator == 1:
            return str(int(self.exponent)).translate(SUPERSCRIPT_MAP)

        # Returns fractional representation as ¹/²
        return f"{numerator}/{denominator}".translate(SUPERSCRIPT_MAP)

    @property
    def name(self) -> str:
        """ Constructs the dimension's name using symbols. """
        return self._name_cache

    @classmethod
    def dimensionless(cls) -> Dimension:
        """Factory method for dimensionless. """
        return cls(FBase.DIMENSIONLESS, 1)

    def __str__(self) -> str:
        """ Returns name for __str__ dunder method. """
        return self.name

    def __repr__(self) -> str:
        """ Displays the defined dimension through its components. """
        return f"<Dimension name='{self.name}'>"


# Mapping int to unicode for __repr__ & name within Dimension
SUPERSCRIPT_MAP = str.maketrans("/0123456789+-. ", "⁄⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻· ")

# Standard SI Secondary fallbacks (used when not overridden in preferences)
_SIBASE_SYMBOLS = {
    FBase.TIME: "s",
    FBase.LENGTH: "m",
    FBase.MASS: "kg",
    FBase.CURRENT: "A",
    FBase.TEMPERATURE: "K",
    FBase.AMOUNT: "mol",
    FBase.LUMINOSITY: "cd",
    FBase.DIMENSIONLESS: "∅",
}

# Preferred order for consistent dimensional notation (e.g., kg·m²/s²)
# Mass first (0), then length (1), time (2), etc.
_ORDER = {
    FBase.MASS: 0,
    FBase.LENGTH: 1,
    FBase.TIME: 2,
    FBase.CURRENT: 3,
    FBase.TEMPERATURE: 4,
    FBase.AMOUNT: 5,
    FBase.LUMINOSITY: 6,
    FBase.DIMENSIONLESS: 7,
}



"""
Filename: lazy_imports.py
Author: William Bowley

Description:
    This file defines a generic lazy import
    helper function and a specific import for
    factory as its the main lazy import.
"""


from typing import Any
from functools import lru_cache


class LazyImportError(ImportError):
    """ Exception for failed lazy imports """
    def __init__(self, caller: str, module: str):
        """ Returns a custom error message for lazy imports """
        msg = (
            f"Could not import '{module}' for '{caller}'. "
            "This usually means picounits was not installed correctly"
        )
        super().__init__(msg)


@lru_cache(maxsize=None)
def import_factory(caller_name: str) -> Any:
    """ Caches/returns the 'Factory' import route for lazy imports """
    try:
        from picounits.core.quantities.factory import Factory
        return Factory

    except ImportError:
        raise LazyImportError(caller_name, "Factory") from None


@lru_cache(maxsize=None)
def lazy_import(
    module_path: str, method_name: str, caller_name: str
) -> Any:
    """ 
    Caches/returns the module_path.module_name for lazy imports 
    """
    try:
        mod = __import__(module_path, fromlist=[method_name])
        return getattr(mod, method_name)

    except ImportError:
        raise LazyImportError(caller_name, method_name) from None



"""
Filename: equations.py
Author: William Bowley
Version: 0.1

Description:
    Equations for a analytical coil-gun model
    as an example, for the picounits library
"""

from math import pi, ceil
from picounits.core import unit_validator, Quantity as q

from picounits.constants import (
    CURRENT, VOLTAGE, FLUX_DENSITY, FORCE, TIME, MASS, LENGTH,
    INDUCTANCE, IMPEDANCE, DIMENSIONLESS, PERMEABILITY
)


@unit_validator(MASS)
def projectile_mass(axial_length: q, radius: q, density: q) -> q:
    """ Calculates the mass of the projectile """
    volume = pi * radius ** 2 * axial_length
    return volume * density


@unit_validator(DIMENSIONLESS)
def estimate_turns(
    axial_length: q, inner_radius: q, outer_radius: q,
    wire_diameter: q, fill_factor: q
) -> q:
    """ Estimates the number of turns that can fit in a rectangular coil """
    slot_area = axial_length * (outer_radius - inner_radius)
    wire_area = wire_diameter ** 2
    effective_area = slot_area * fill_factor

    turns = effective_area / wire_area
    return ceil(turns)


@unit_validator(PERMEABILITY)
def calculate_approximate_core_permeability(
    position: q, permeability: q, relative: q, coil_length: q,
    proj_length: q, coil_outer_radius: q, proj_radius: q
) -> q:
    """ Calculates the approximate permeability as the coil moves through """
    # Calculates the cross sectional axial-area
    coil_area = 2 * coil_outer_radius * coil_length

    # Calculates the rectangular bounding boxes of projectile and coil
    px_min, py_min = position - proj_length, -proj_radius
    px_max, py_max = position, +proj_radius

    cx_min, cy_min = 0 * LENGTH, -coil_outer_radius
    cx_max, cy_max = coil_length, +coil_outer_radius

    # Calculates the width and height of the overlap
    overlap_x = max(0 * LENGTH, min(px_max, cx_max) - max(px_min, cx_min))
    overlap_y = max(0 * LENGTH, min(py_max, cy_max) - max(py_min, cy_min))

    # Calculates the intersection area and ratio
    in_area = overlap_x * overlap_y
    occupancy = in_area / coil_area

    # Restricts to 0 -> 1
    occupancy = min(1.0, max(0.0, occupancy))

    # Returns weighted permeability of the core
    return permeability * (1 + (relative - 1) * occupancy)


@unit_validator(IMPEDANCE)
def cal_resistance(
    turns: q, mean_radius: q, wire_diameter: q, resistivity: q
) -> q:
    """ Calculates the resistance of the inductor """
    length = turns * pi * mean_radius
    area = pi * (wire_diameter / 2) ** 2
    return resistivity * length / area


@unit_validator(INDUCTANCE)
def cal_inductance(
    turns: q, axial_length: q, mean_radius: q, permeability: q
) -> q:
    """ Calculates the inductance of solenoid """
    # Defines the permeability of free space
    area = pi * mean_radius ** 2
    return (turns ** 2 * area * permeability) / axial_length


@unit_validator(FORCE)
def projectile_drag(velocity: q, density: q, coefficient: q, radius: q) -> q:
    """ Calculates the drag force on the projectile """
    area = pi * radius ** 2
    velocity = velocity * abs(velocity)
    drag_force = -0.5 * density * velocity * area * coefficient
    return drag_force


@unit_validator(CURRENT / TIME)
def differential_currents(voltage: q, inductance: q) -> q:
    """ Differential equation for current within the system """
    return voltage / inductance


@unit_validator(CURRENT)
def clipping_current(current_limit: q, current: q) -> q:
    """ Limits the current to simulate a current limiting supply"""
    return min(current_limit, current)


@unit_validator(VOLTAGE)
def inductor_voltage(
    supply_voltage: q, current: q, resistance: q, induced_voltage: q
) -> q:
    """ Computes the inductor voltage during simulation"""
    voltage_drop = current * resistance
    return supply_voltage - voltage_drop + induced_voltage


@unit_validator(CURRENT)
def rk_2nd_order_current(
    current: q, voltage: q, inductance: q, resistance: q, step_size: q
) -> q:
    """ Solves the differential equation using Ralston's method"""
    k1 = differential_currents(voltage, inductance)

    # Updates the voltage for the next predicted frame
    voltage = voltage - resistance * 3 / 4 * step_size * k1
    k2 = differential_currents(voltage, inductance)

    current += (1 / 3 * k1 + 2 / 3 * k2) * step_size
    return current


@unit_validator(FLUX_DENSITY)
def position_b_field(
    position: q, current: q, turns: q, coil_length: q,
    coil_radius: q, relative_permeability: q, saturation: q
) -> q:
    """
    Calculates the axial B-field inside a solenoid based on projectile position
    """
    # print(position, current, turns, coil_length, coil_radius)
    b_constant = 0.5 * relative_permeability * turns * current

    denom1 = (position ** 2 + coil_radius ** 2) ** 0.5
    term1 = position / denom1

    delta = position - coil_length
    denom2 = (delta ** 2 + coil_radius ** 2) ** 0.5
    term2 = delta / denom2

    b_field = b_constant * (term1 - term2)

    # Simple switch to mimic B-H curve
    if abs(b_field) > saturation:
        return saturation * (b_field / abs(b_field))

    return b_field


@unit_validator(FORCE)
def inst_force(
    b_field: q, permeability: q, projectile_outer_radius: q, direction: q
) -> q:
    """
    Calculate instantaneous force on a ferromagnetic projectile.

    Based on energy density in a magnetic field:
    F ≈ (1/2) * (B^2 / µ) * A, where A is the cross-sectional area.
    """
    cross_section = pi * projectile_outer_radius ** 2
    force_magnitude = (
        0.5 * (abs(b_field) * b_field / permeability) * cross_section
    )

    return force_magnitude * direction



"""
Filename: arithmetic.py
Author: William Bowley
Version: 0.2

Description:
    Defines the methods for operation instance methods 
    for vector quantities
"""

from numpy import ndarray, dot as np_dot, cross as np_cross, arccos

from picounits.core.unit import Unit

from picounits.core.quantities.packet import Packet
from picounits.core.quantities.factory import Factory


def _temporary_numpy_requirement(q1: Packet, q2: Packet) -> None:
    """
    NOTE: This method should be replaced with a conversion layer
    that handles multiple vector representations (numpy, symbolic, etc.)
    """
    if not isinstance(q1.value, ndarray) or not isinstance(q2.value, ndarray):
        msg = (
            f"Vector operations currently only support numpy arrays. "
            f"Got {type(q1.value).__name__} and {type(q2.value).__name__}. "
            f"Cannot perform operation between {q1} and {q2}"
        )
        raise NotImplementedError(msg)


def dot(q1: Packet, q2: Packet) -> Packet:
    """ Computes the dot product of two vector packets """
    Factory.category_check(q1, q2)

    # NOTE: This method should be replaced with a conversion layer
    _temporary_numpy_requirement(q1, q2)

    # Computes the dot product of values and multiples units
    result_value = np_dot(q1.value, q2.value)
    result_unit = q1.unit * q2.unit

    return Factory.create(result_value, result_unit)


def cross(q1: Packet, q2: Packet) -> Packet:
    """ Computes the cross product of two vector packets """
    Factory.category_check(q1, q2)

    # NOTE: This method should be replaced with a conversion layer
    _temporary_numpy_requirement(q1, q2)

    if q1.value.size != 3 or q2.value.size != 3:
        msg = "Cross product is only defined for 3D vectors"
        raise ValueError(msg)

    # Compute cross product of values
    result_value = np_cross(q1.value, q2.value)

    # Compute resulting unit (multiplication of units)
    result_unit = q1.unit * q2.unit
    return Factory.create(result_value, result_unit)

def angle_between(q1: Packet, q2: Packet) -> Packet:
    """ Computes the angle between q1 and q2 vectors. Returns radians """
    Factory.category_check(q1, q2)

    # NOTE: This method should be replaced with a conversion layer
    _temporary_numpy_requirement(q1, q2)

    # Check units are compatible
    q1.unit_check(q2)

    # Compute magnitudes
    q1_mag = q1.magnitude
    q2_mag = q2.magnitude

    if q1_mag == 0 or q2_mag == 0:
        msg = "Cannot compute angle with zero-magnitude vector"
        raise ValueError(msg)

    # Compute dot product and normalize by magnitudes
    dot_product = np_dot(q1.value, q2.value)
    cos_angle = dot_product / (q1_mag * q2_mag)

    # Clamp to [-1, 1] to handle floating point errors
    cos_angle = max(-1.0, min(1.0, cos_angle))

    angle_rad = arccos(cos_angle)
    return Factory.create(angle_rad, Unit.dimensionless())


def normalize(q1: Packet) -> Packet:
    """ Returns a unit vector in the same direction as this vector """
    mag = q1.magnitude

    if mag == 0:
        msg = "Cannot normalize a zero-magnitude vector"
        raise ValueError(msg)

    # Divide value by magnitude to get unit vector
    normalized_value = q1.value / mag
    return Factory.create(normalized_value, q1.unit)



"""
Filename: scales.py
Author: William Bowley
Version: 0.6
Clear: Y

Description
    The prefix scale 'PrefixScale' is encoded as a enum
    to enable human readable scales for when displaying
    Quantities to the user interface.
"""

from __future__ import annotations
from typing import Any
from enum import Enum

from picounits.lazy_imports import import_factory, lazy_import

class PrefixScale(Enum):
    """
    Prefix Scale for Quantities:

    Represent the possible prefix scales for quantities.
    Supports normalization of values and scaling
    """
    YOTTA   = 24
    ZETTA   = 21
    EXA     = 18
    PETA    = 15
    TERA    = 12
    GIGA    = 9
    MEGA    = 6
    KILO    = 3
    BASE    = 0
    CENTI   = -2
    MILLI   = -3
    MICRO   = -6
    NANO    = -9
    PICO    = -12
    FEMTO   = -15
    ATTO    = -18
    ZEPTO   = -21
    YOCTO   = -24

    @classmethod
    def from_value(cls, power: int) -> PrefixScale:
        """ Return closest PrefixScale; on ties, prefer smaller scale """
        if not isinstance(power, int):
            msg = f"Power must be an int, not {type(power)}"
            raise TypeError(msg)

        min_distance, closest_member = float('inf'), None

        # O(n) iteration
        for member in cls:
            distance = abs(member.value - power)

            # Update if closer, or tie with smaller scale
            if distance < min_distance or (
                distance == min_distance
                and closest_member is not None
                and member.value < closest_member.value
            ):
                min_distance, closest_member = distance, member

            # Early exit on exact match
            if distance == 0:
                break

        # Fail safe (unlikely): If fails to assign a member, raise error
        if closest_member is None:
            msg = f"No suitable PrefixScale found for power: {power}"
            raise ValueError(msg)

        return closest_member

    @classmethod
    def from_symbol(cls, reference: str) -> PrefixScale | None:
        """
        Compares reference symbol with symbol lookup, if not found returns none
        """
        if not isinstance(reference, str):
            return None

        return _SYMBOLS_TO_SCALE.get(reference)

    @classmethod
    def all_symbols(cls) -> list[str]:
        """ Returns a list of all symbols """
        return list(_SYMBOLS_TO_SCALE.items())

    @property
    def symbol(self) -> str:
        """ Returns the prefix symbol. """
        return _SCALE_SYMBOLS[self]

    def __str__(self) -> str:
        """ Returns name for __str__ dunder method """
        return self.symbol

    def __repr__(self) -> str:
        """ Displays the scale name and its power of 10 """
        return f"<PrefixScale.{self}: 10^{self.value}>"

    def __rmul__(self, other: Any):
        """
        Acts as a syntactic bridge to Quantity to allow for a cleaner API.
        Example: 10 * kilo * LENGTH => Quantity(10, LENGTH, kilo)
        """
        factory = import_factory('PrefixScale.__rmul__')
        unit = lazy_import(
            "picounits.core.unit", "Unit", 'PrefixScale.__rmul__'
        )

        return factory.create(other, unit(), self)


# Fast mapping for enums and ensure o(1) lookup
# Scale symbols dictionary only uses ASCII to ensure easy usage in .uiv files
_SCALE_SYMBOLS = {
    PrefixScale.YOTTA:  "Y",
    PrefixScale.ZETTA:  "Z",
    PrefixScale.EXA:    "E",
    PrefixScale.PETA:   "P",
    PrefixScale.TERA:   "T",
    PrefixScale.GIGA:   "G",
    PrefixScale.MEGA:   "M",
    PrefixScale.KILO:   "k",
    PrefixScale.BASE:   "",
    PrefixScale.CENTI:  "c",
    PrefixScale.MILLI:  "m",
    PrefixScale.MICRO:  "u",
    PrefixScale.NANO:   "n",
    PrefixScale.PICO:   "p",
    PrefixScale.FEMTO:  "f",
    PrefixScale.ATTO:   "a",
    PrefixScale.ZEPTO:  "z",
    PrefixScale.YOCTO:  "y",
}

# Generates a reverse lookup table to ensure o(1) lookup
_SYMBOLS_TO_SCALE = {symbol: scale for scale, symbol in _SCALE_SYMBOLS.items()}



"""
Filename: operations.py
Author: William Bowley

Description:
    Defines the mathematical operators for
    usage by the converter when parsing .uiv format.
"""

from __future__ import annotations
from enum import Enum, auto

from picounits.extensions.parser_errors import ParserError

class Operations(Enum):
    """ Map of Mathematical Operations for Units """
    MULTIPLICATION = auto()
    POWER = auto()
    DIVIDED = auto()

    @property
    def symbol(self) -> str:
        """ Returns the operation symbol via direct lookup """
        return _LOOKUP_OPERATORS.get(self)[0]

    @property
    def _repr_name(self) -> str:
        """ Returns the operator object name """
        return f"<Operations type={self.name}, symbol={self.symbol}>"

    @classmethod
    def from_symbol(cls, char: str) -> Operations:
        """ Creation of operator object via symbol direct lookup """
        operator = _LOOKUP_STRINGS.get(char)
        if operator is None:
            ops = list(_LOOKUP_STRINGS.keys())
            msg = f"'{char}' is an unknown operator. Supported operators are {ops}"
            raise ParserError(cls.__name__, msg)

        return operator

    @classmethod
    def check_unicode_power(cls, power: str) -> int | False:
        """ Returns converted unicode power """
        try:
            return SUPERSCRIPT_MAP[power]
        except KeyError:
            return False

    @classmethod
    def validate_unicode_usage(cls, tokens: list[str]) -> None:
        """ Ensures non mixing of ^ with unicode superscripts """
        for index, token in enumerate(tokens):
            if token == "^" and index + 1 < len(tokens):
                if tokens[index+1] in SUPERSCRIPT_MAP:
                    msg = (
                        f"Ambiguous power syntax: '^' followed by unicode "
                        f"'{tokens[index+1]}'"
                    )
                    raise ParserError(cls.__name__, msg)

    @classmethod
    def all_symbols(cls) -> list[str]:
        """ Returns a list of all symbols """
        return list(_LOOKUP_STRINGS.keys())

    def __repr__(self) -> str:
        """ Return string representation for terminal"""
        return self._repr_name

    def __str__(self) -> str:
        """ Return string representation """
        return self._repr_name


# Direct lookup table for string notation
_LOOKUP_STRINGS = {
    "*": Operations.MULTIPLICATION,
    "x": Operations.MULTIPLICATION,
    "·": Operations.MULTIPLICATION,
    "∙": Operations.MULTIPLICATION,
    "/": Operations.DIVIDED,
    "÷": Operations.DIVIDED,
    "^": Operations.POWER,
    "⁰": Operations.POWER,
    "¹": Operations.POWER,
    "²": Operations.POWER,
    "³": Operations.POWER,
    "⁴": Operations.POWER,
    "⁵": Operations.POWER,
    "⁶": Operations.POWER,
    "⁷": Operations.POWER,
    "⁸": Operations.POWER,
    "⁹": Operations.POWER,
}

# Unicode lookup table
SUPERSCRIPT_MAP = {
    "⁰": 0,
    "¹": 1,
    "²": 2,
    "³": 3,
    "⁴": 4,
    "⁵": 5,
    "⁶": 6,
    "⁷": 7,
    "⁸": 8,
    "⁹": 9,
}

# Direct lookup table for operator enum to string notation
_LOOKUP_OPERATORS = {}
for symbol, op in _LOOKUP_STRINGS.items():
    _LOOKUP_OPERATORS.setdefault(op, []).append(symbol)



"""
Filename: parser.py
Author: William Bowley

Description:
    High-level parser for .uiv (Unit-Informed Values) files.
    Orchestrates tokenization, conversion, and construction.
"""

from __future__ import annotations
from typing import Any
from pathlib import Path


from picounits.extensions.utilities.tokenizer import Tokenizer
from picounits.extensions.utilities.construction import Construct
from picounits.extensions.utilities.converter import Converter
from picounits.extensions.parser_errors import ParserError

from picounits.configuration.config import get_derived_units

from picounits.extensions.loader import DynamicLoader
from picounits.core.unit import Unit


class Parser:
    """ Parser for .uiv (unit informed values) file format """

    @classmethod
    def _back_compatibility(cls, status: bool, filepath: str) -> None:
        """ Displays a back compatibility message to the user """
        if status:
            return

        print(
            f"Tip: {Path(filepath).name} lacks 'format' key; "
            "use 'version.format: 0.1.0' for better compatibility."
        )

    @classmethod
    def _unit_frame_compatibility(cls, status: bool, filepath: str) -> None:
        """ Display a unit frame compatibility message to the user """
        if status:
            return

        print(
            f"Tip: {Path(filepath).name} lacks 'unit_frame' key; "
            "use 'version.unit_frame: '(your_derived_units).ut' for better compatibility."
        )

    @classmethod
    def _is_section(cls, line: str) -> tuple[bool, str | None]:
        """ Check if line defines a section [name] """
        line = line.strip()
        if line.startswith('[') and line.endswith(']'):
            return True, line[1:-1]

        return False, None

    @classmethod
    def _should_skip(cls, line: str) -> bool:
        """ Check if line should be skipped (empty or comment) """
        line = line.strip()
        return not line or line.startswith('#')

    @classmethod
    def _extract_qualities(
        cls, value_str: str
    ) -> tuple[Any, str | list, str | list]:
        """ Extracts value, prefix and unit from raw string. """
        value_str = value_str.strip()

        # Check for quoted string
        if Tokenizer.is_quoted(value_str):
            return Tokenizer.strip_quotes(value_str), "", ""

        # Check for list structure
        if value_str.startswith('['):
            bracket_result = Tokenizer.extract_bracket_content(value_str, 0)
            if not bracket_result:
                msg = f"Invalid list structure: {value_str!r}"
                raise ParserError(cls.__name__, msg)

            content, end_index = bracket_result
            list_value = Converter.parse_list(f"[{content}]")
            remainder = value_str[end_index + 1:].strip()

            if remainder:
                # Extract unit specifications
                units = Tokenizer.extract_paren_groups(remainder)

                if len(units) > 1:
                    # Multiple units (column-wise)
                    prefixes = []
                    for item in remainder.split(','):
                        paren_idx = item.find('(')
                        prefix = item[:paren_idx].strip()
                        prefixes.append(prefix)

                    return list_value, prefixes, units

                elif len(units) == 1:
                    # Single unit with potential prefix
                    unit = units[0]

                    # Check for prefix before unit
                    last_paren = remainder.rfind('(')
                    if last_paren > 0:
                        before = remainder[:last_paren].rstrip()

                        if before and before[-1].isalpha():
                            rest = before[:-1].rstrip()

                            if not rest or rest[-1] in ',)':
                                return list_value, before[-1], unit

                    return list_value, "", unit

            return list_value, "", ""

        # Single value with potential unit
        paren_groups = Tokenizer.extract_paren_groups(value_str)

        if paren_groups:
            # Find where unit starts
            paren_idx = value_str.find('(')
            head = value_str[:paren_idx].strip()
            unit = paren_groups[0]

            # Check for prefix (single letter before unit, not part of number)
            if head and not head[-1].isdigit() and not head.endswith('.'):
                value = head[:-1].strip()
                prefix = head[-1]
            else:
                value = head
                prefix = ""

            return Converter.cast(value), prefix, unit

        # Just a plain value with no unit
        return Converter.cast(value_str), "", ""

    @classmethod
    def _parse_lines(cls, lines: list[str], filepath: str, derived_file: str) -> dict:
        """ Extract logic from raw_open into reusable method """
        data = {}
        current_status = False
        current_section = None

        # Compatibility status checks
        format_status = False
        unit_frame_status = False
        if derived_file is None:
            # Due to no derived units being used hence no compatibility issues
            unit_frame_status = True

        i = 0
        while i < len(lines):
            line = lines[i].strip()
            i += 1

            # Skip comments and empty lines
            if cls._should_skip(line):
                continue

            # Check for section
            is_section, section_name = cls._is_section(line)
            if is_section:
                current_section = section_name
                data[section_name] = {}

                if section_name.lower() == "version":
                    current_status = True
                else:
                    current_status = False

                continue

            # Parse key-value pair
            split_result = Tokenizer.split_key_value_pairs(line)
            if not split_result:
                continue

            if current_section is None:
                msg = f"Key-value pair outside section: {line!r}"
                raise ParserError(cls.__name__, msg)

            key, raw_value = split_result

            # Handle multi-line values (lists that span multiple lines)
            if raw_value.startswith('['):
                open_count, close_count = Tokenizer.count_brackets(raw_value)

                # Collect lines until balanced
                while open_count > close_count and i < len(lines):
                    next_line = lines[i].strip()
                    i += 1
                    raw_value += ' ' + next_line

                    next_open, next_close = Tokenizer.count_brackets(next_line)
                    open_count += next_open
                    close_count += next_close

            # Extract value, prefix, and unit
            value, prefix, unit = cls._extract_qualities(raw_value)

            # Checks for 'format & unit frame' to verify back_compatibility
            if current_status:
                if key.lower() == "format":
                    format_status = True

                if key.lower() == "unit_frame" and unit_frame_status is False:
                    if value != derived_file:
                        msg = (
                            f"Failed to load {filepath!r} due to unit frame mismatch "
                            f"{value!r} != {derived_file}"
                        )
                        raise ValueError(msg)

                    unit_frame_status = True

            # Constructs array of quantity (value_1: unit_1, ..., value_n: unit_n)
            if isinstance(unit, list) and isinstance(value, list):
                valid = False
                for entry in value:
                    if isinstance(entry, (complex, int, float)):
                        valid = True
                    else:
                        valid = False

                if valid:
                    array = []
                    for index, value in enumerate(value):
                        quantity = Construct.quantity(value, prefix[index], unit[index])
                        array.append(quantity)

                    data[current_section][key] = array
                    continue

            # Construct the quantity (Arrays, single value unit pairs)
            quantity = Construct.quantity(value, prefix, unit)
            data[current_section][key] = quantity

        cls._back_compatibility(format_status, filepath)
        cls._unit_frame_compatibility(unit_frame_status, filepath)
        return data

    @classmethod
    def open(
        cls, filepath_or_file, derived_units=None, loader_class=None
    ) -> DynamicLoader:
        """ Parse .uiv file into structured data via attribute injection. """
        if loader_class is None:
            loader_class = DynamicLoader

        # Loads .ut derived units
        if derived_units:
            _, derived_file = get_derived_units(Path(derived_units))
        else:
            _, derived_file = get_derived_units()

        # If it's a file-like object, read lines directly
        if hasattr(filepath_or_file, "read"):
            lines = filepath_or_file.readlines()
            return loader_class(cls._parse_lines(lines, filepath_or_file, derived_file))

        else:
            with open(filepath_or_file, 'r', encoding='utf-8') as f:
                lines = f.readlines()
                return loader_class(cls._parse_lines(lines, filepath_or_file, derived_file))

    @classmethod
    def open_derived(cls, filepath) -> dict[str, Unit]:
        """ Parses a units.uiv file into a symbol -> Unit registry """
        with open(filepath, 'r', encoding='utf-8') as f:
            lines = f.readlines()

        status = False
        registry = {}
        for line in lines:
            if cls._should_skip(line):
                continue

            result = Tokenizer.split_key_value_pairs(line.strip())
            if not result:
                continue

            symbol, unit_str = result

            if symbol.lower() == "format":
                status = True
                continue

            registry[symbol] = Construct.tokenize_unit(unit_str)

        cls._back_compatibility(status, filepath)
        return registry



"""
Filename: construction.py
Author: William Bowley

Description:
    Handles unit and quantity construction 
    via tokenization of the dimensional algebra.
    This allows for reconstruction of the Unit,
    which is used to construct the quantity.
"""

from typing import Any

from picounits.core.unit import Unit
from picounits.core.scales import PrefixScale
from picounits.core.dimensions import FBase, Dimension
from picounits.core.quantities.packet import Packet
from picounits.core.quantities.factory import Factory

from picounits.extensions.parser_errors import ParserError
from picounits.extensions.utilities.operations import Operations

from picounits.configuration.config import get_derived_units

class Construct:
    """ Stateless construction utility for unit and quantity construction """

    @classmethod
    def _construct_from_tokenized_unit(cls, tokens: list[str]) -> 'Unit':
        """ Constructs the unit from already tokenized unit string """
        if not tokens:
            return Unit.dimensionless()

        result = Unit.dimensionless()
        queue_op = None
        pending_unit: Unit | None = None
        pending_power: int | None = None

        # Validate unicode usage & imports of derived units
        Operations.validate_unicode_usage(tokens)
        _effective_derived, _ = get_derived_units()

        for token in tokens:
            symbol = FBase.from_symbol(token)
            if symbol:
                pending_unit = Unit(Dimension(symbol))
                continue

            # Uses a lookup table for custom unit types.
            if token in _effective_derived:
                pending_unit = _effective_derived[token]
                continue

            # Not a base symbol -> try operation
            try:
                operation = Operations.from_symbol(token)
            except ParserError:
                try:
                    exp = int(token)
                    if pending_unit is not None:
                        pending_unit = pending_unit ** exp
                        pending_power = None   # clear
                    else:
                        msg = f"Stray number '{token}'"
                        raise ParserError(cls.__name__, msg) from None
                    continue

                except ValueError:
                    pass

                # Special case: unicode superscript (whole token)
                exp = Operations.check_unicode_power(token)
                if exp is not None:
                    if pending_unit is not None:
                        pending_unit = pending_unit ** exp
                    continue

                msg = f"Unknown token '{token}'"
                raise ParserError(cls.__name__, msg) from None

            if operation is Operations.POWER:
                # (power will be consumed in the next iteration)
                continue

            # Apply pending unit before changing operator
            if pending_unit is not None:
                if queue_op == Operations.MULTIPLICATION:
                    result *= pending_unit
                elif queue_op == Operations.DIVIDED:
                    result /= pending_unit
                elif queue_op is None:
                    result = pending_unit
                pending_unit = None

            queue_op = operation

        # Last dimension operations
        if pending_unit is not None:
            if pending_power is not None:
                pending_unit **= pending_power
            if queue_op == Operations.MULTIPLICATION:
                result *= pending_unit
            elif queue_op == Operations.DIVIDED:
                result /= pending_unit
            elif queue_op is None:
                result = pending_unit

        return result

    @classmethod
    def tokenize_unit(cls, unit_str: str) -> Unit:
        """ Constructs unit from parsed unit strings """
        if not unit_str:
            # Handles dimensionless values
            return Unit.dimensionless()

        # Splits the unit string into tokenized between operators
        for op in Operations.all_symbols():
            unit_str = unit_str.replace(op, f" {op} ")

        tokens = unit_str.split()

        if len(tokens) == 1:
            # Handles case of a single dimension within the unit
            dim = FBase.from_symbol(tokens[0])

            if not dim:
                try:
                    _effective_derived, _ = get_derived_units()
                    return _effective_derived[tokens[0]]

                except KeyError:
                    symbols = FBase.all_symbols()
                    msg = (
                        f"'{tokens[0]}' is an unknown dimension. "
                        f"Supported dimensions are {symbols}"
                    )
                    raise ParserError(cls.__name__, msg) from None

            return Unit(Dimension(dim))

        return cls._construct_from_tokenized_unit(tokens)

    @classmethod
    def _construct_prefix(cls, prefix: str) -> PrefixScale:
        """ Constructs a prefix via the o(1) prefix lookup """
        prefix_str = prefix if isinstance(prefix, str) else ""
        prefix_scale = PrefixScale.from_symbol(prefix_str)

        if not prefix_scale:
            if prefix_str:
                prefixes = PrefixScale.all_symbols()
                msg = (
                    f"'{prefix_str}' is an unknown prefix. "
                    f"Supported prefixes are {prefixes}"
                )
                raise ParserError(cls.__name__, msg)

            prefix_scale = PrefixScale.BASE

        return prefix_scale

    @classmethod
    def _nested_arrays(
        cls, value: Any, prefix: str | list, unit: str | list
    ) -> Packet:
        """ Constructs a nested array into a nested quality array"""
        result = []
        for sublist in value:
            row = []
            for index, value in enumerate(sublist):
                # Finds the unit for that columns data
                if index < len(unit):
                    column_unit = unit[index]
                else:
                    column_unit = unit[-1]

                # Finds the prefix for that columns data
                if isinstance(prefix, list):
                    if index < len(prefix):
                        column_prefix = prefix[index]
                    else:
                        column_prefix = prefix[-1]
                elif isinstance(prefix, str):
                    column_prefix = prefix
                else:
                    column_prefix = ""

                # Builds the nested array as a array of quantities
                row.append(cls.quantity(value, column_prefix, column_unit))
            result.append(row)
        return result

    @classmethod
    def _quantity_arrays(
        cls, value: Any, prefix: str | list, unit: str | list
    ) -> Packet:
        """ Constructs a series of packets or uses a packet array """
        # Nested array check (list of lists)
        if value and isinstance(value[0], list):
            # For nested arrays with multiple units
            if isinstance(unit, list) and len(unit) > 1:
                # Apply different unit to each column
                return cls._nested_arrays(value, prefix, unit,)

        # Simple lists, create array quantity e.x [10, 2] (kg*m*s^-2)
        prefix = Construct._construct_prefix(prefix)

        unit_str = unit[0] if isinstance(unit, list) else unit
        unit_obj = cls.tokenize_unit(unit_str)

        return Factory.create(value, unit_obj, prefix)

    @classmethod
    def quantity(
        cls, value: Any, prefix: str | list, unit: str | list
    ) -> Packet:
        """ Constructs a quantity object from prefix, value, unit """
        if isinstance(value, (str, bool)):
            # Handles non-numerical values as non-quantities
            return value

        if isinstance(value, list):
            # Handles simple lists and nested arrays of quantities
            return cls._quantity_arrays(value, prefix, unit)

        # Lookups prefix via symbol
        prefix = cls._construct_prefix(prefix)

        # Finds and creates the unit
        unit_str = unit if isinstance(unit, str) else ""
        unit_obj = cls.tokenize_unit(unit_str)

        return Factory.create(value, unit_obj, prefix)



"""
Filename: factory.pyi
Author: William Bowley
Version: 0.1
Clear: X

Description:
    Static type hinting for the factory module.
    The factory is type caster for quantities and
    dispatcher for mathematical methods
"""

from numpy import ndarray, integer, floating, complexfloating

from typing import Callable, overload
from picounits.core.unit import Unit
from picounits.core.scales import PrefixScale
from picounits.core.quantities.scalars.types.complex import ComplexPacket
from picounits.core.quantities.scalars.types.real import RealPacket
from picounits.core.quantities.vectors.types.array import ArrayPacket


class Factory:
    @overload
    @classmethod
    def create(
        cls, value: complex | complexfloating, unit: Unit, prefix: PrefixScale | None = None
    ) -> ComplexPacket: ...

    @overload
    @classmethod
    def create(
        cls, value: float | int | floating | integer, prefix: PrefixScale | None = None
    ) -> RealPacket: ...

    @overload
    @classmethod
    def create(
        cls, value: ndarray | list | tuple, prefix: PrefixScale | None = None
    ) -> ArrayPacket: ...

    @classmethod
    def create(
        cls, 
        value: complex | float | int | floating | ndarray | list | tuple | integer | complexfloating,
        unit: Unit,
        prefix: PrefixScale | None = None,
    ) -> ComplexPacket | RealPacket | ArrayPacket: ...

    @classmethod
    def reallocate(cls, op_name: str) -> Callable[[Callable], Callable]: ... 
    
    @classmethod
    def category_check(cls, q1: Packet, q2: Packet) -> None: ...


"""
Filename: dimensional_construction.py
Author: William Bowley

Description:
    Tests for the fundamental unit 'FBASE' and dataclass 'Dimension'
    to ensure that the dimensional construction 
    is correct for the application.
    
    NOTE:
    Testing for string representation should be added.
"""

import unittest
from picounits.core.dimensions import Dimension, FBase, SUPERSCRIPT_MAP


class DimensionConstruction(unittest.TestCase):
    """ Tests dimensional construction system"""
    def test_construct_dimensions(self):
        """ Construct a dimension with each known base unit """
        for base in FBase:
            _ = Dimension(base, 1)

    def test_construction_with_non_Fbase(self):
        """ Non-valid construction without Fbase """
        invalid_cases = ["MASS", [FBase.AMOUNT], {}, 10, 0.1]

        for base in invalid_cases:
            with self.assertRaises(TypeError):
                Dimension(base, 1)

    def test_construction_with_non_numerical_exponent(self):
        """ Non-valid construction with non float or integer exponent """
        invalid_cases = [
            (FBase.AMOUNT, "Im non-numerical"),
            (FBase.MASS, [10, 10, 10]),
            (FBase.LENGTH, {})
        ]

        for base, exponent in invalid_cases:
            with self.assertRaises(TypeError):
                Dimension(base, exponent)

    def test_construct_dimensionless(self):
        """ Construct of dimensionless via x^0=1 or redefining exponent to 1 """
        zero_exponent = Dimension(FBase.LENGTH, 0)
        redefined_to_one = Dimension(FBase.DIMENSIONLESS, 10)

        # Zero exponent should return base of dimensionless and exponent of 1
        self.assertIs(zero_exponent.base, FBase.DIMENSIONLESS)
        self.assertEqual(zero_exponent.exponent, 1)

        # Redefined to one should mutate exponent from 10 to 1
        self.assertEqual(redefined_to_one.exponent, 1)

    def test_integer_float_and_fractional_superscripts(self):
        """ Construct of dimensions with integer, float and fractional superscripts """
        def _convert_to_unicode(superscript: str) -> str:
            """ Converts to unicode from ascii """
            return superscript.translate(SUPERSCRIPT_MAP)

        base_cases = [
            (1, "1"), (1/2, "1/2"), (1/18, "1/18"), (-1, "-1"), (-1/5, "-1/5"),
            (0.1, "1/10"), (0.05, "1/20"), (10, "10"), (1/2500, "1/2500"), (2.0, "2")
        ]

        for exponent, expected in base_cases:
            dim = Dimension(FBase.MASS, exponent)
            superscript = dim.superscript

            self.assertEqual(superscript, _convert_to_unicode(expected))

    def test_dimensionless_factory(self):
        """ Factory method for dimensionless outputs """
        dim = Dimension.dimensionless()
        self.assertEqual(dim, Dimension(FBase.DIMENSIONLESS, 1))

if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
Filename: cli.py
Author: William Bowley
Version: 0.1
Clean: N

Description:
    Simple command line tool to generate the '.picounits'
    automatically to working directories
"""

from __future__ import annotations

import argparse
from pathlib import Path

from picounits.configuration.picounits import DEFAULT_CONFIG


def generate(args: argparse.Namespace | None = None) -> None:
    """ Generates the '.picounits' file in working directories """
    _ = args
    target = Path.cwd() / ".picounits"
    if target.exists():
        print(f"Warning: .picounits already exists at {target}")
        reply = input("Overwrite? (y/N): ").strip().lower()

        if reply != "y":
            print("Aborted. No changes made.")
            return None

    try:
        target.write_text(DEFAULT_CONFIG.strip() + "\n", encoding="utf-8")
        print(f"Successfully created .picounits at:\n   {target}")
        print("\n You can now edit it to switch to custom symbols (t/l/m)")
        print(" or change the dimension order.")
        print(
            " picounits will automatically use your settings in this project!"
        )
    except OSError as e:
        print(f"Failed to write .picounits: {e}")
        return

    # Offer to show the content
    reply = input("   Show the generated config now? (Y/n): ").strip().lower()
    if reply != "n":
        print("\n--- Generated .picounits content ---")
        print(DEFAULT_CONFIG)
        print("------------------------------------")


def main():
    """ Adds the argparse argument """
    parser = argparse.ArgumentParser(
        prog="picounits",
        description="picounits — flexible, project-aware units for Python"
    )
    subparsers = parser.add_subparsers()

    gen_parser = subparsers.add_parser(
        "generate",
        help=(
            "Create a default .picounits config file in the current directory"
        ),
        description=(
            "Generate a ready-to-use .picounits file with helpful"
            "comments " "and both SI defaults and common alternatives."
        )
    )

    gen_parser.set_defaults(func=generate)

    args = parser.parse_args()
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()

# Default configuration file for .picounits



"""
Filename: parser_errors.py
Author: William Bowley

Description:
    Holds the parser error class to ensure
    no circular imports occur during runtime.
"""


class ParserError(ValueError):
    """ Exception for parser errors when parsing """
    def __init__(self, caller: str, error: str):
        """ Returns a custom error message """
        msg = f"'{caller}' raised error: {error}. "
        super().__init__(msg)



"""
Filename: transcendental.py
Author: William Bowley
Version: 0.3

Description:
    Defines the methods for transcendental mathematical
    functions methods for scalar quantities (REAL only)
"""

from math import (
    radians, degrees, sin, cos, tan, exp, log, log10, log2, e,
    asin, acos, atan, atan2, sinh, cosh, tanh, asinh, acosh, atanh
)

from picounits.core.unit import Unit
from picounits.core.quantities.packet import Packet
from picounits.core.quantities.factory import Factory


def _valid_input_for_transcendental(q: Packet, method: str) -> None:
    """ if unit is not dimensionless, it raises a value error """
    if q.unit == Unit.dimensionless():
        return None

    msg = (
        f"Method '{method}' requires dimensionless Quantity, "
        f"{q.unit} != {Unit.dimensionless()}"
    )
    raise ValueError(msg)


def _valid_input_for_inverse(q: Packet, method: str) -> None:
    """ If -1 > q.value or q.value > 1, raises a value error """
    if -1 <= q.value <= 1:
        return

    m = q.value
    msg = f"Method '{method}' requires input in range [-1, 1], got {m}"
    raise ValueError(msg)


def _valid_input_for_inverse_reciprocal(q: Packet, method: str) -> None:
    """ If |q.value| < 1, raises a value error (for asec, acsc) """
    if abs(q.value) >= 1:
        return

    m = q.value
    msg = f"Method '{method}' requires |x| >= 1, got {m}"
    raise ValueError(msg)


def _valid_input_for_logarithms(
    q: Packet, n: float | int, method: str
) -> None:
    """ If q.value <= 0 or n <= 0 or n == 1, raises value error """
    if not isinstance(n, (float, int)):
        mag = type(q.value)
        msg = f"Method '{method}' requires a float or int, got {mag}"
        raise TypeError(msg)

    if q.value <= 0:
        msg = f"Method '{method}' requires magnitude > 0, got {q.value}"
        raise ValueError(msg)

    if n <= 0 or n == 1:
        msg = f"Method '{method}' requires "
        msg += "n != 1" if n == 1 else "n > 0"
        msg += f", got {n}"
        raise ValueError(msg)


""" ANGLE CONVERSION """


def to_radians_logic(q: Packet) -> Packet:
    """ Defines the logic for converting to radians with Quantities"""
    _valid_input_for_transcendental(q, to_radians_logic.__name__)
    new_magnitude = radians(q.value)

    return Factory.create(new_magnitude, q.unit)


def to_degrees_logic(q: Packet) -> Packet:
    """ Defines the logic for converting to degrees with Quantities """
    _valid_input_for_transcendental(q, to_degrees_logic.__name__)
    new_magnitude = degrees(q.value)

    return Factory.create(new_magnitude, q.unit)


""" BASIC TRIGONOMETRIC FUNCTIONS """


def sin_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the sine operation on a Quantity """
    _valid_input_for_transcendental(q, sin_logic.__name__)
    new_magnitude = sin(q.value)

    return Factory.create(new_magnitude, q.unit)


def cos_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the cosine operation on a Quantity """
    _valid_input_for_transcendental(q, cos_logic.__name__)
    new_magnitude = cos(q.value)

    return Factory.create(new_magnitude, q.unit)


def tan_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the tangent operation on a Quantity
    """
    _valid_input_for_transcendental(q, tan_logic.__name__)
    new_magnitude = tan(q.value)

    return Factory.create(new_magnitude, q.unit)


def csc_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the cosecant operation on a Quantity
    """
    _valid_input_for_transcendental(q, csc_logic.__name__)
    new_magnitude = 1 / sin(q.value)

    return Factory.create(new_magnitude, q.unit)


def sec_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the secant operation on a Quantity
    """
    _valid_input_for_transcendental(q, sec_logic.__name__)
    new_magnitude = 1 / cos(q.value)

    return Factory.create(new_magnitude, q.unit)


def cot_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the cotangent operation on a Quantity
    """
    _valid_input_for_transcendental(q, cot_logic.__name__)
    new_magnitude = 1 / tan(q.value)

    return Factory.create(new_magnitude, q.unit)


""" INVERSE TRIGONOMETRIC FUNCTIONS """


def asin_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc sine operation on a Quantity
    """
    _valid_input_for_transcendental(q, asin_logic.__name__)
    _valid_input_for_inverse(q, asin_logic.__name__)
    new_magnitude = asin(q.value)

    return Factory.create(new_magnitude, q.unit)


def acos_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc cosine operation on a Quantity
    """
    _valid_input_for_transcendental(q, acos_logic.__name__)
    _valid_input_for_inverse(q, acos_logic.__name__)
    new_magnitude = acos(q.value)

    return Factory.create(new_magnitude, q.unit)


def atan_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc tan operation on a Quantity
    """
    _valid_input_for_transcendental(q, atan_logic.__name__)
    new_magnitude = atan(q.value)

    return Factory.create(new_magnitude, q.unit)


def atan2_logic(q_y: Packet, q_x: Packet) -> Packet:
    """
    Defines the logic for performing the arc tan 2 operation on a Quantity
    """
    # Ensures both q_y and q_x have the same unit
    q_y.unit_check(q_x)

    # Calculates the new magnitude, unit is always dimensionless
    new_magnitude = atan2(q_y.value, q_x.value)

    return Factory.create(new_magnitude, Unit.dimensionless())


def acsc_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc cosecant operation on a Quantity
    """
    _valid_input_for_transcendental(q, acsc_logic.__name__)
    _valid_input_for_inverse_reciprocal(q, acsc_logic.__name__)

    new_magnitude = asin(1 / q.value)
    return Factory.create(new_magnitude, q.unit)


def asec_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc secant operation on a Quantity
    """
    _valid_input_for_transcendental(q, asec_logic.__name__)
    _valid_input_for_inverse_reciprocal(q, asec_logic.__name__)

    new_magnitude = acos(1 / q.value)
    return Factory.create(new_magnitude, q.unit)


def acot_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the arc cotangent operation on a Quantity
    """
    _valid_input_for_transcendental(q, acot_logic.__name__)

    """
    Specific function check: If magnitude == 0, raises a ZeroDivisionError
    """
    if q.value == 0:
        mag = q.value
        msg = f"Method '{acot_logic.__name__}' requires q != 0, got {mag}"
        raise ZeroDivisionError(msg)

    new_magnitude = atan(1 / q.value)

    return Factory.create(new_magnitude, q.unit)


""" HYPERBOLIC FUNCTIONS """


def sinh_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic sine operation """
    _valid_input_for_transcendental(q, sinh_logic.__name__)
    new_magnitude = sinh(q.value)

    return Factory.create(new_magnitude, q.unit)


def cosh_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic cosine operation """
    _valid_input_for_transcendental(q, cosh_logic.__name__)
    new_magnitude = cosh(q.value)

    return Factory.create(new_magnitude, q.unit)


def tanh_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic tangent operation """
    _valid_input_for_transcendental(q, tanh_logic.__name__)
    new_magnitude = tanh(q.value)

    return Factory.create(new_magnitude, q.unit)


def csch_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic cosecant operation """
    _valid_input_for_transcendental(q, csch_logic.__name__)
    new_magnitude = 1 / sinh(q.value)

    return Factory.create(new_magnitude, q.unit)


def sech_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic secant operation """
    _valid_input_for_transcendental(q, sech_logic.__name__)
    new_magnitude = 1 / cosh(q.value)

    return Factory.create(new_magnitude, q.unit)


def coth_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the hyperbolic cotangent operation """
    _valid_input_for_transcendental(q, coth_logic.__name__)
    new_magnitude = 1 / tanh(q.value)

    return Factory.create(new_magnitude, q.unit)


""" INVERSE HYPERBOLIC FUNCTIONS """


def asinh_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic sine operation
    """
    _valid_input_for_transcendental(q, asinh_logic.__name__)
    new_magnitude = asinh(q.value)

    return Factory.create(new_magnitude, q.unit)


def acosh_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic cosine operation
    """
    _valid_input_for_transcendental(q, acosh_logic.__name__)

    """
    Specific function check: If magnitude < 1, raises a value error
    """
    if q.value < 1:
        mag = q.value
        msg = f"Method '{acosh_logic.__name__}' requires q >= 1, got {mag}"
        raise ValueError(msg)

    new_magnitude = acosh(q.value)

    return Factory.create(new_magnitude, q.unit)


def atanh_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic tangent operation
    """
    _valid_input_for_transcendental(q, atanh_logic.__name__)

    """
    Specific function check: If -1 >= magnitude or magnitude >= 1,
    raises a value error
    """
    if -1 >= q.value or q.value >= 1:
        mag = q.value
        msg = f"Method '{atanh_logic.__name__}' requires -1 < q < 1, got {mag}"
        raise ValueError(msg)

    new_magnitude = atanh(q.value)

    return Factory.create(new_magnitude, q.unit)


def acsch_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic cosecant operation
    """
    _valid_input_for_transcendental(q, acsch_logic.__name__)

    """
    Specific function check: If magnitude equals 0, raises a value error
    """
    if q.value == 0:
        mag = q.value
        msg = f"Method '{acsch_logic.__name__}' requires q != 0, got {mag}"
        raise ValueError(msg)

    new_magnitude = asinh(1 / q.value)
    return Factory.create(new_magnitude, q.unit)


def asech_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic secant operation
    """
    _valid_input_for_transcendental(q, asech_logic.__name__)

    """
    Specific function check: If magnitude < 0 or magnitude > 1,
    raise a value error
    """
    if q.value > 1 or q.value <= 0:
        mag = q.value
        msg = (
            f"Method '{asech_logic.__name__}' requires 0 < q <= 1, got {mag}"
        )
        raise ValueError(msg)

    new_magnitude = acosh(1 / q.value)
    return Factory.create(new_magnitude, q.unit)


def acoth_logic(q: Packet) -> Packet:
    """
    Defines the logic for performing the inverse hyperbolic cotangent operation
    """
    _valid_input_for_transcendental(q, acoth_logic.__name__)

    """ Specific function check: If |magnitude| <= 1, raise value error """
    if abs(q.value) <= 1:
        mag = q.value
        msg = f"Method '{acoth_logic.__name__}' requires |q| > 1, got {mag}"
        raise ValueError(msg)

    new_magnitude = atanh(1 / q.value)
    return Factory.create(new_magnitude, q.unit)


""" EXPONENTIAL AND LOGARITHMIC FUNCTIONS """


def exp_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the exponential operation """
    _valid_input_for_transcendental(q, exp_logic.__name__)
    new_magnitude = exp(q.value)

    return Factory.create(new_magnitude, q.unit)


def log_logic(q: Packet, n: float | int) -> Packet:
    """ Defines the logic for performing the variable log """
    _valid_input_for_transcendental(q, log_logic.__name__)
    _valid_input_for_logarithms(q, n, log_logic.__name__)

    new_magnitude = log(q.value, n)
    return Factory.create(new_magnitude, q.unit)


def log2_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the log 2 """
    _valid_input_for_transcendental(q, log2_logic.__name__)
    _valid_input_for_logarithms(q, 2, log2_logic.__name__)

    new_magnitude = log2(q.value)
    return Factory.create(new_magnitude, q.unit)


def log10_logic(q: Packet) -> Packet:
    """ Defines the logic for performing the log 10 """
    _valid_input_for_transcendental(q, log10_logic.__name__)
    _valid_input_for_logarithms(q, 10, log10_logic.__name__)

    new_magnitude = log10(q.value)
    return Factory.create(new_magnitude, q.unit)


def nlog_logic(q: Packet) -> Packet:
    """" Defines the logic for performing the natural logarithm """
    _valid_input_for_transcendental(q, nlog_logic.__name__)
    _valid_input_for_logarithms(q, e, nlog_logic.__name__)

    new_magnitude = log(q.value)
    return Factory.create(new_magnitude, q.unit)



"""
Filename: validators.py
Author: William Bowley
Version: 0.2
Clear: X

Description:
    Defines the methods for the unit validator
    method in the Packet ABC dataclass
"""

from typing import Callable

from picounits.core.unit import Unit
from picounits.core.quantities.packet import Packet

# Easy reference for users
Quantity = Packet


class DimensionError(ValueError):
    """ Exception for unit error """
    def __init__(self, caller: str, message: str):
        """ Returns a custom error message for unit error """
        msg = f"{caller!r} {message}"
        super().__init__(msg)


def _check_packet(q: Packet, func: str) -> None:
    """ Checks to ensure q is a packet """
    if isinstance(q, Packet):
        return

    msg = f"returned {type(q)}, expected packet"
    raise DimensionError(func, msg)


def _check_forecasted(u1: Unit, u2: Unit, func: str) -> None:
    """ Checks to ensure u1 is equal to u2 """
    if isinstance(u1, Unit) and isinstance(u2, Unit):
        if u1 == u2:
            return

    msg = f"returned {u1}, expected {u2}"
    raise DimensionError(func, msg)


def unit_validator(forecasted: Unit) -> Callable:
    """
    A decorator; it checks a function unit output, if different raises an error
    """
    def decorator(func) -> Callable:
        def wrapper(*args, **kwargs) -> Callable:
            result = func(*args, **kwargs)

            # Single Packet
            if isinstance(result, Packet):
                _check_forecasted(result.unit, forecasted, func.__name__)
                return result

            # Tuple or list of Packets
            if isinstance(result, (tuple, list)):
                for item in result:
                    _check_packet(item, func.__name__)
                    _check_forecasted(
                        item.unit, forecasted, func.__name__
                    )
                return result

            msg = (
                f"{func.__name__} returned {type(result)}, "
                "expected Packet or tuple/list of Packets"
            )
            raise TypeError(msg)

        return wrapper
    return decorator



"""
Filename: dimensional_algebra.py
Author: William Bowley

Description:
    Tests for the Unit class to ensure that the 
    dimensional algebra and pipelines are correct.
    
    NOTE:
    Testing for string representation should be added.
"""

import unittest
from picounits.core.dimensions import Dimension, FBase
from picounits.core.unit import Unit

# Defining fundamental dimensions

_TIME           = Dimension(base=FBase.TIME)
_LENGTH         = Dimension(base=FBase.LENGTH)
_MASS           = Dimension(base=FBase.MASS)
_CURRENT        = Dimension(base=FBase.CURRENT)
_TEMPERATURE    = Dimension(base=FBase.TEMPERATURE)
_AMOUNT         = Dimension(base=FBase.AMOUNT)
_LUMINOSITY     = Dimension(base=FBase.LUMINOSITY)
_DIMENSIONLESS  = Dimension(base=FBase.DIMENSIONLESS)


class DimensionAlgebra(unittest.TestCase):
    """ Tests dimensional algebra system """
    def test_construct_units(self):
        """ Constructs each fundamental unit """
        fundamental = []
        for base in FBase:
            fundamental.append(Dimension(base))

        for dim in fundamental:
            _ = Unit(dim)

    def test_non_Dimension_construction(self):
        """ Tests invalid construction via using float, int, etc """
        cases = ["mass", 10.1, 1, -1, {}, []]

        for dim in cases:
            with self.assertRaises(ValueError):
                _ = Unit(dim)

    def test_removal_of_duplicated_bases(self):
        """ Removal of duplicated bases """
        mass = Dimension(FBase.MASS)

        with self.assertRaises(ValueError):
            _ = Unit(mass, mass)

    def test_removal_of_dimensionless(self):
        """ Tests removal of dimensionless """
        length = Dimension(FBase.LENGTH)
        dimensionless = Dimension.dimensionless()

        unit = Unit(length, dimensionless)
        self.assertEqual(unit.dimensions, [length])

    def test_dimensionless_factory(self):
        """ Factory method for dimensionless outputs """
        dimensionless_via_factory = Unit.dimensionless()
        dimensionless_via_construction = Unit(_DIMENSIONLESS)

        self.assertEqual(
            dimensionless_via_construction, dimensionless_via_factory
        )

    def test_unit_forwards_multiplication(self):
        """ Tests multiplication different units together """
        cases = [
            ((Unit(_LENGTH), Unit(_TIME)), Unit(_LENGTH, _TIME)),
            ((Unit(_MASS), Unit(_TIME)), Unit(_MASS, _TIME)),
            ((Unit(_CURRENT), Unit(_MASS)), Unit(_CURRENT, _MASS)),
            ((Unit(_AMOUNT), Unit(_LUMINOSITY)), Unit(_LUMINOSITY, _AMOUNT)),
            (
                (Unit(_MASS), Unit(Dimension(FBase.CURRENT, 2))),
                Unit(_MASS, Dimension(FBase.CURRENT, 2))
            ),
            (
                (Unit(_MASS), Unit(_MASS)), Unit(Dimension(FBase.MASS, 2))
            )
        ]

        for case in cases:
            units, expected = case
            result = units[0] * units[1]

            self.assertEqual(result.dimensions, expected.dimensions)

    def test_unit_forwards_true_division(self):
        """ Tests true division between different units """
        cases = [
            (
                (Unit(_DIMENSIONLESS), Unit(_TEMPERATURE)),
                Unit(Dimension(FBase.TEMPERATURE, -1))
            ),
            (
                (Unit(_TEMPERATURE), Unit(_AMOUNT)),
                Unit(Dimension(FBase.AMOUNT, -1), _TEMPERATURE)
            ),
            (
                (Unit(_MASS), Unit(Dimension(FBase.CURRENT, 2))),
                Unit(_MASS, Dimension(FBase.CURRENT, -2))
            ),
            (
                (1, Unit(_MASS)), # Reciprocal of the unit (edge case)
                Unit(Dimension(FBase.MASS, -1))
            )
        ]

        for case in cases:
            units, expected = case
            results = units[0] / units[1]

            self.assertEqual(results.dimensions, expected.dimensions)

    def test_unit_backwards_true_division(self):
        """ Tests true division between a unit and different value """
        with self.assertRaises(ValueError):
            _ = Unit(_MASS) / 10

    def test_unit_forwards_power(self):
        """ Tests power between a unit and a integer or float """
        cases = [
            ((Unit(_LENGTH), 1), Unit(_LENGTH)),
            ((Unit(_MASS), -2), Unit(Dimension(FBase.MASS, -2))),
            ((Unit(_LENGTH), 10.0), Unit(Dimension(FBase.LENGTH, 10))),
            ((Unit(_CURRENT), 0.1), Unit(Dimension(FBase.CURRENT, 0.1))),
            ((Unit(_TIME), -0.1), Unit(Dimension(FBase.TIME, -0.1)))
        ]

        for case in cases:
            units, expected = case
            results = units[0] ** units[1]

            self.assertEqual(results.dimensions, expected.dimensions)

    def test_unit_backwards_power(self):
        """ Tests backwards power between a number and a unit """
        with self.assertRaises(TypeError):
            _ = 1 ** Unit(_TEMPERATURE)

if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
Filename: constants.py
Author: William Bowley

Description:
    Defines semantic physical quantities and their
    dimensional relationships in picounits.
"""

from picounits.core.unit import Unit
from picounits.core.dimensions import FBase, Dimension
from picounits.core.scales import PrefixScale

""" =============== Predefined scales for quantities =============== """

GIGA                    = PrefixScale.GIGA
MEGA                    = PrefixScale.MEGA
KILO                    = PrefixScale.KILO
CENTI                   = PrefixScale.CENTI
MILLI                   = PrefixScale.MILLI
MICRO                   = PrefixScale.MICRO
NANO                    = PrefixScale.NANO
PICO                    = PrefixScale.PICO


""" =============== Fundamental dimensions =============== """

TIME                    = Unit(Dimension(base=FBase.TIME))
LENGTH                  = Unit(Dimension(base=FBase.LENGTH))
MASS                    = Unit(Dimension(base=FBase.MASS))
CURRENT                 = Unit(Dimension(base=FBase.CURRENT))
TEMPERATURE             = Unit(Dimension(base=FBase.TEMPERATURE))
AMOUNT                  = Unit(Dimension(base=FBase.AMOUNT))
LUMINOSITY              = Unit(Dimension(base=FBase.LUMINOSITY))
DIMENSIONLESS           = Unit(Dimension(base=FBase.DIMENSIONLESS))
NULLSET                  = DIMENSIONLESS

""" =============== Geometric quantities =============== """

AREA                    = LENGTH ** 2
VOLUME                  = LENGTH ** 3


""" =============== Kinematics =============== """

DISPLACEMENT            = LENGTH
DISTANCE                = LENGTH
VELOCITY                = LENGTH / TIME
SPEED                   = VELOCITY
ACCELERATION            = LENGTH / TIME ** 2
FREQUENCY               = DIMENSIONLESS / TIME
PERIOD                  = TIME
WAVENUMBER              = DIMENSIONLESS / LENGTH
ANGULAR_FREQUENCY       = DIMENSIONLESS / TIME
PHASE                   = DIMENSIONLESS


""" =============== Classical mechanics =============== """

FORCE                   = MASS * LENGTH / TIME ** 2
MOMENTUM                = MASS * VELOCITY
ANGULAR_MOMENTUM        = MASS * LENGTH ** 2 / TIME
TORQUE                  = FORCE * LENGTH
ENERGY                  = MASS * LENGTH ** 2 / TIME ** 2
POWER                   = ENERGY / TIME
PRESSURE                = FORCE / AREA
DENSITY                 = MASS / VOLUME
WEIGHT                  = FORCE


""" =============== Thermodynamics =============== """

ENTROPY                     = ENERGY / TEMPERATURE
HEAT_CAPACITY               = ENERGY / TEMPERATURE
SPECIFIC_HEAT               = ENERGY / (MASS * TEMPERATURE)
THERMAL_CONDUCTIVITY        = POWER / (LENGTH * TEMPERATURE)
CONVECTION_COEFFICIENT      = POWER / (LENGTH**2 * TEMPERATURE)
VOLUMETRIC_HEAT_CAPACITY    = HEAT_CAPACITY / VOLUME
VOLUMETRIC_HEATING          = POWER / LENGTH ** 3
DIFFUSIVITY                 = LENGTH**2 * TIME**-1

""" =============== Electromagnetism =============== """

CHARGE                  = CURRENT * TIME
ELECTRIC_FIELD          = FORCE / CHARGE
ELECTRIC_POTENTIAL      = ENERGY / CHARGE
VOLTAGE                 = ELECTRIC_POTENTIAL
RESISTANCE              = VOLTAGE / CURRENT
CONDUCTANCE             = DIMENSIONLESS / RESISTANCE
CAPACITANCE             = CHARGE / VOLTAGE
IMPEDANCE               = ENERGY / (TIME * CURRENT ** 2)
INDUCTANCE              = IMPEDANCE * TIME
MAGNETIC_FIELD          = FORCE / (CURRENT * LENGTH)
MAGNETIC_FLUX           = MAGNETIC_FIELD * AREA
PERMEABILITY            = INDUCTANCE / LENGTH
FLUX_DENSITY            = MASS / (TIME ** 2 * CURRENT)
COERCIVITY              = CURRENT / LENGTH
CONDUCTIVITY            = MASS ** -1 * LENGTH ** -3 * TIME ** 3 * CURRENT ** 2


""" =============== Waves & radiation =============== """

INTENSITY               = POWER / AREA
LUMINANCE               = LUMINOSITY / AREA
RADIANT_FLUX            = POWER


""" =============== Dimensionless semantic quantities =============== """

STRAIN                  = DIMENSIONLESS
REFRACTIVE_INDEX        = DIMENSIONLESS
EFFICIENCY              = DIMENSIONLESS
COEFFICIENT             = DIMENSIONLESS
PROBABILITY             = DIMENSIONLESS



""" picounits/__init__.py """

from typing import Any

from picounits.constants import *
from picounits.core.quantities.validators import (
    unit_validator, Quantity
)

Q = Quantity


class UnitError(TypeError):
    """ Exception for Unit Error """
    def __init__(self, error: str):
        """ Returns a custom error message """
        msg = f"raised error: {error}. "
        super().__init__(msg)


def check_quantity(quantity: Quantity, ref: Quantity) -> None:
    """ Checks if the quantity has the correct reference unit """
    if not isinstance(quantity, Quantity):
        msg = f"{type(quantity)!r} is not a physical quantity object"
        raise UnitError(msg)

    if not isinstance(ref, (Unit, Quantity)):
        msg = f"Reference unit must be either a quantity or unit, not {type(ref)}"
        raise UnitError(msg)

    if isinstance(ref, Quantity):
        if quantity.unit != ref.unit:
            msg = f"Expected {ref.unit!r}, got {quantity.unit!r}"
            raise UnitError(msg)

    if isinstance(ref, Unit):
        if quantity.unit != ref:
            msg = f"Expected {ref!r}, got {quantity.unit!r}"
            raise UnitError(msg)


def strip_quantity(quantity: Quantity, reference: Quantity) -> Any:
    """ Strips quantity from value returns raw value """
    check_quantity(quantity, reference)

    return quantity.value



"""
File: runner.py
Author: William Bowley

Description:
    Main script to run all unit test modules within
    the picounits library. This includes unit modelling, 
    parser, dynamic-loader and configurations.
"""

import unittest

from unit_test.unit.dimensional_algebra import DimensionAlgebra
from unit_test.unit.dimensional_construction import DimensionConstruction


loader = unittest.TestLoader()
suite = unittest.TestSuite()

suite.addTests(loader.loadTestsFromTestCase(DimensionConstruction))
suite.addTests(loader.loadTestsFromTestCase(DimensionAlgebra))


runner = unittest.TextTestRunner(verbosity=2)

if __name__ == "__main__":
    result = runner.run(suite)



"""
Filename: tokenizer.py
Author: William Bowley

Description:
    Low-level tokenization for .uiv format.
    Handles string parsing, bracket matching,
    delimiter splitting.
"""

from __future__ import annotations
from picounits.extensions.parser_errors import ParserError

class Tokenizer:
    """ Stateless tokenization utilities for parsing """

    @classmethod
    def is_quoted(cls, text: str) -> bool:
        """ Check if text is surrounded by quotes """
        text = text.strip()

        if len(text) < 2:
            # Empty quoted text Ex. ("")
            return False

        if text.startswith('"') and text.endswith('"'):
            # Start and end with double quotes
            return True

        elif text.startswith("'") and text.endswith("'"):
            # Start and end with single quotes
            return True

        return False

    @classmethod
    def strip_quotes(cls, text: str) -> str:
        """ Removes surrounding quotes if present """
        text = text.strip()
        if cls.is_quoted(text):
            # Removes double quotation
            return text[1:-1]
        return text

    @classmethod
    def count_brackets(cls, text: str) -> tuple[int, int]:
        """ Count opening and closing brackets """
        return text.count('['), text.count(']')

    @classmethod
    def split_key_value_pairs(cls, line: str) -> tuple[str, str] | None:
        """ Split line on first unquoted colon Ex. speed: 10 M(ms^-1)"""
        quote_character = None

        for index, character in enumerate(line):
            # Handles if character is escaped
            if index > 0 and line[index-1] == '\\':
                continue

            # Handles ':' inside quotation via toggle logic
            if character in "\"'":
                # Ex. "Hello" true inside 'hello' none outside.
                if not quote_character:
                    quote_character = character
                elif character == quote_character:
                    quote_character = None

            elif character == ':' and not quote_character:
                return line[:index].strip(), line[index+1:].strip()

        return None

    @classmethod
    def extract_bracket_content(
        cls, text: str, start_index: int = 0
    ) -> tuple[str, int] | None:
        """
        Extracts content between matching brackets Ex. '[1,2,3]' -> ('1,2,3', 6)
        """
        open_index = text.find('[', start_index)

        if open_index == -1:
            # Handles when '[' is not found
            return None

        bracket_pair_depth = 0
        quote_character = None
        for index in range(open_index, len(text)):
            character = text[index]

            # Handles if character is escaped
            if index > open_index and text[index-1] == '\\':
                continue

            # Handles brackets inside quotes
            if character in "\"'":
                # Ex. "Hello" true inside 'hello' none outside.
                if not quote_character:
                    quote_character = character
                elif character == quote_character:
                    quote_character = None

            # skip bracket depth check if inside quotes
            if quote_character:
                continue

            if text[index] == '[':
                bracket_pair_depth += 1

            elif text[index] == ']':
                bracket_pair_depth -= 1
                if bracket_pair_depth == 0:
                    # Found matching brackets
                    content = text[open_index + 1:index]
                    return content, index

        msg = f"{text!r} requires balanced brackets to be parsed."
        raise ParserError(cls.__name__, msg)

    @classmethod
    def extract_paren_groups(cls, text: str) -> list[str]:
        """ Extracts all top-level parenthesized groups. """
        groups = []
        index = 0

        while index < len(text):
            # Skip non-paren characters
            while index < len(text) and text[index] not in '(':
                index += 1

            if index >= len(text):
                break

            # Found opening paren - track depth
            depth = 1
            start = index + 1
            index += 1
            quote_character = None

            while index < len(text) and depth > 0:
                character = text[index]

                # Handle escaped characters
                if index > 0 and text[index-1] == '\\':
                    index += 1
                    continue

                # Handle quotes
                if character in "\"'":
                    if not quote_character:
                        quote_character = character
                    elif character == quote_character:
                        quote_character = None

                # Handle parens (only if not in quotes)
                elif not quote_character:
                    if character == '(':
                        depth += 1
                    elif character == ')':
                        depth -= 1
                        if depth == 0:
                            # Found matching close paren
                            groups.append(text[start:index])
                            index += 1
                            break

                index += 1

            # Check for unclosed paren
            if depth > 0:
                msg = f"Unbalanced parentheses in: {text}"
                raise ParserError(cls.__name__, msg)

        return groups


"""
Filename: arithmetic.py
Author: William Bowley
Version: 0.2

Description:
    Defines the methods for arithmetic dunder
    and instance methods for scalar quantities
"""


from picounits.core.unit import Unit
from picounits.constants import DIMENSIONLESS

from picounits.core.quantities.packet import Packet
from picounits.core.quantities.factory import Factory


@Factory.reallocate("__add__")
def add_logic(q1: Packet, q2: Packet) -> Packet:
    """ Defines the logic for adding two packets """

    # Additions requires both packets to have the same Unit
    q1.unit_check(q2)

    # Calculates the new value and returns new packet
    new_value = q1.value + q2.value
    return Factory.create(new_value, q1.unit)


@Factory.reallocate("__sub__")
def sub_logic(q1: Packet, q2: Packet) -> Packet:
    """ Defines the logic for adding subtracting two packets """

    # Subtraction requires both packets to have the same Unit
    q1.unit_check(q2)

    # Calculates the new value and returns new packet
    new_value = q1.value - q2.value
    return Factory.create(new_value, q1.unit)


@Factory.reallocate("__mul__")
def multiplication_logic(q1: Packet, q2: Packet | Unit) -> Packet:
    """ Defines the logic for multiplication between two quantities"""

    if not isinstance(q2, Unit):
        # Calculates the new value and unit, then returns new packet
        new_value = q1.value * q2.value
        new_unit = q1.unit * q2.unit

        return Factory.create(new_value, new_unit)

    # Creates a syntactic bridge for defining quantities with units
    # Ex. 10 * Kilo = Q(10000, DIMENSIONLESS) -> Q * TIME -> Q(10000, s)
    if q1.unit == DIMENSIONLESS:
        return Factory.create(q1.value, q2)

    msg = f"Failed to perform multiplication between {q1} and {q2}"
    raise RuntimeError(msg)


@Factory.reallocate("__truediv__")
def true_division_logic(q1: Packet, q2: Packet) -> Packet:
    """ Defines the logic for true division between two quantities """
    if q2.value == 0:
        m1, m2 = q1.value, q2.value  # inline length to long
        msg = f'True Division failed due to division by zero: {m1} / {m2}'
        raise ValueError(msg)

    # Calculates new value and unit, than returns packet
    new_value = q1.value / q2.value
    new_unit = q1.unit / q2.unit

    return Factory.create(new_value, new_unit)


@Factory.reallocate("__floordiv__")
def floor_division_logic(q1: Packet, q2: Packet) -> Packet:
    """ Defines the logic for floor division between two quantities """
    if q2.value == 0:
        msg = f'Floor Division failed due to division by zero: {q1.value} // {q2.value}'
        raise ValueError(msg)

    # Performs floor division (//) on values and standard division (/) on units
    new_value = q1.value // q2.value
    new_unit = q1.unit / q2.unit

    return Factory.create(new_value, new_unit)


@Factory.reallocate("__pow__")
def power_logic(q1: Packet, q2: Packet) -> Packet:
    """ Defines the logic for power between two quantities """
    if q2.unit != DIMENSIONLESS:
        msg = f"A unit cannot be raised to the power of a unit; {q1}^{q2}"
        raise TypeError(msg)

    # Handles zero exponent rule (x^0 = 1)
    if q2.value == 0:
        return Factory.create(1.0, DIMENSIONLESS)

    # Calculates new value and unit, than returns packet
    new_value = q1.value ** q2.value
    new_unit = q1.unit ** q2.value

    return Factory.create(new_value, new_unit)



"""
Filename: units.py
Author: William Bowley
Clear: Y

Description:
    Defines the Unit class which is comprised
    of a Dimension dataclass. This class performs
    abstract dimensional analysis using Dimensions.
"""

from __future__ import annotations
from typing import Any

from picounits.core.dimensions import Dimension
from picounits.lazy_imports import import_factory, lazy_import


class Unit:
    """
    A Physical Unit: A series of different Dimension together

    Allows for abstract dimensional analysis and complex unit
    representation for displaying to the user interface
    """
    # Uses __slots__ to decrease memory overhead per object
    __slots__ = ('dimensions', '_hash_cache', '_name_cache')

    def __init__(self, *dimensions: Dimension) -> None:
        """ Initialize the unit; assume dimensionless if no dimensions are given """
        if not dimensions:
            dimensions = [Dimension.dimensionless()]

        # Handles units defined with non-Dimension's
        for dim in dimensions:
            if isinstance(dim, Dimension):
                continue

            msg = f"Dimensions must be 'Dimension' not {type(dim).__name__}"
            raise ValueError(msg)

        self.dimensions = list(dimensions)
        self._hash_cache = None
        self._name_cache = None

        # Performs checks for consistent representation
        self._duplicated_bases_check()
        self._remove_dimensionless()
        self._sort_order()

    def _remove_dimensionless(self) -> None:
        """ If more than one dimension, remove dimensionless """
        if self.length == 1:
            return

        new_dimensions = []
        for dimension in self.dimensions:
            if dimension != Dimension.dimensionless():
                new_dimensions.append(dimension)

        self.dimensions = new_dimensions

    def _duplicated_bases_check(self) -> None:
        """ Checks for duplicated bases during initialization """
        duplicated_bases = set()

        for dim in self.dimensions:
            if dim.base not in duplicated_bases:
                duplicated_bases.add(dim.base)
                continue

            msg = f"Cannot define a unit with duplicated bases: {dim.base}"
            raise ValueError(msg)

    def _sort_order(self) -> None:
        """ Sorts the dimensions to ensure canonical representation """
        self.dimensions.sort(key=lambda d: d.base.order)

    def _dimensional_analysis(self, other: Unit, division: bool) -> Unit:
        """ Combines or divides two units via their dimension exponents. """
        combined: dict = {dim.base: dim.exponent for dim in self.dimensions}
        for dim in other.dimensions:
            key = dim.base

            # Exponent rules via product and quotient exponent rule
            exponent_change = dim.exponent * (-1 if division else 1)

            if key in combined:
                combined[key] += exponent_change
            else:
                combined[key] = exponent_change

        # Reconstruct dimensions, filtering out zero exponents (dimensionless)
        new_dimensions: list[Dimension] = []
        for base, exponent in combined.items():
            if exponent != 0:
                new_dimensions.append(Dimension(base, exponent))

        # Handles dimensionless unit if all dimensions canceled out
        if not new_dimensions:
            return Unit()

        return Unit(*new_dimensions)

    @property
    def name(self) -> str:
        """ Returns the units name as dimensions """
        if self._name_cache is None:
            get_derived_units = lazy_import(
                "picounits.configuration.config", "get_derived_units", "Unit.name"
            )
            derived, _ = get_derived_units()

            # Primary: Check exact match first
            for symbol, unit in derived.items():
                if self.dimensions == unit.dimensions:
                    self._name_cache = symbol
                    return self._name_cache

            # Secondary: Try partial substitution
            remaining = list(self.dimensions)
            result_parts = []

            for symbol, unit in derived.items():
                derived_dims = list(unit.dimensions)
                if all(d in remaining for d in derived_dims):
                    # Remove matched dimensions from remaining
                    for d in derived_dims:
                        remaining.remove(d)

                    result_parts.append(symbol)

            # Fallback: Append non-substituted dimensions
            result_parts.extend(str(d.name) for d in remaining)
            self._name_cache = "·".join(result_parts)

        return self._name_cache

    @property
    def length(self) -> int:
        """ Number of distinct dimension bases (e.g., kg·m·s⁻² has length 3) """
        return len(self.dimensions)

    @classmethod
    def dimensionless(cls) -> Dimension:
        """Factory method for dimensionless. """
        # Note an empty unit is automatically dimensionless, this exists to
        # ensure that the API is explicit instead of implicit with just unit()
        dimensionless = Dimension.dimensionless()
        return cls(dimensionless)

    def __mul__(self, other: Unit) -> Unit:
        """ Defines behavior for the forward multiplication operator """
        if isinstance(other, Unit):
            return self._dimensional_analysis(other, False)

        return self.__rmul__(other)

    def __rmul__(self, other: Any):
        """
        Acts as a syntactic bridge to packet to allow for a cleaner API.
        Example: 10 * CURRENT => RealPacket(10, CURRENT)
        """
        factory = import_factory('Unit.__rmul__')
        return factory.create(other, self)

    def __truediv__(self, other: Unit) -> Unit:
        """ Defines behavior for forward true division """
        if isinstance(other, Unit):
            return self._dimensional_analysis(other, True)

        msg = f"Cannot true divide a 'Unit' by {type(other).__name__}"
        raise ValueError(msg)

    def __rtruediv__(self, other: Any):
        """
        Acts as a syntactic bridge to packet and reciprocal method

        Quantity bridge: 10 / CURRENT = RealPacket(10, a⁻¹)
        Reciprocal method : 1 / CURRENT => Unit(A⁻¹)
        """
        factory = import_factory('Unit.__rtruediv__')

        if not isinstance(other, (int, float)):
            msg = (
                f"Cannot use division method with {type(other).__name__} "
            )
            raise ValueError(msg)

        new_dims = [
            Dimension(dim.base, dim.exponent * -1) for dim in self.dimensions
        ]

        if other == 1:
            # Returns the reciprocal of the unit
            return Unit(*new_dims)

        # Returns a packet with value and reciprocal unit
        return factory.create(other, Unit(*new_dims))

    def __pow__(self, other: int | float) -> Unit:
        """ Defines behavior for forward power method """
        if isinstance(other, (float, int)):
            new_dims = [
                Dimension(dim.base, dim.exponent * other)
                for dim in self.dimensions
            ]
            return Unit(*new_dims)

        msg = f"Exponent must be int or float, not {type(other).__name__}"
        raise TypeError(msg)

    def __rpow__(self, other: Any) -> None:
        """ Defines behavior for the right-hand power """
        msg = f"Cannot raise a {type(other)} to the power of a 'Unit'"
        raise TypeError(msg)

    def __eq__(self, other) -> bool:
        """ Checks equality between units of the same length """
        if not isinstance(other, Unit):
            return False
        if self.length != other.length:
            return False

        return hash(self) == hash(other)

    def __hash__(self) -> int:
        """Hash based on dimensions, order-independent"""
        if self._hash_cache is None:
            dim_tuples = {(d.base, d.exponent) for d in self.dimensions}
            self._hash_cache = hash(frozenset(dim_tuples))

        return self._hash_cache

    def __str__(self) -> str:
        """ returns the unit name as a string"""
        return self.name

    def __repr__(self) -> str:
        """ Displays the unit name """
        return f"<Unit: {self.name}>"



"""
Filename: basic.py
Author: William Bowley

Description:
    SUVAT and kinetic energy script showing
    dimensional correctness via picounits
"""

from picounits import LENGTH, TIME, MASS, VELOCITY, ENERGY, KILO
from picounits.core import unit_validator, Quantity as q

# variables
Initial_Velocity = (10+100j) * VELOCITY
Acceleration = 2.5 * LENGTH / TIME ** 2
Displacement = (10+12j) * KILO * LENGTH
Projectile_Mass = 12 * MASS


@unit_validator(VELOCITY)
def suvat(initial_velocity: q, acceleration: q, distance: q) -> q:
    """ Calculates the velocity due to acceleration """
    square = initial_velocity ** 2 + 2 * acceleration * distance
    return square ** 0.5


@unit_validator(ENERGY)
def kinetic_energy(mass: q, velocity: q) -> q:
    """ Calculates the kinetic energy due to new velocity """
    return 0.5 * mass * velocity ** 2


final_velocity = suvat(Initial_Velocity, Acceleration, Displacement)
kinetic = kinetic_energy(Projectile_Mass, final_velocity)
print(f"Velocity: {final_velocity:.3f}", f"Energy: {kinetic:.3f}")



"""
Filename: real.py
Author: William Bowley
Version: 0.9
Clear: X

Description:
    Defines the Real Packet Class which is
    comprised of a Value, Unit and prefixScale.
"""

from math import log10, ceil, trunc, floor
from typing import Any
from dataclasses import dataclass
from numpy import integer, floating

from picounits.core.unit import Unit
from picounits.core.scales import PrefixScale

from picounits.core.quantities.packet import Packet
from picounits.core.quantities.scalars.scalar import ScalarPacket

from picounits.lazy_imports import import_factory
from picounits.configuration.picounits import STANDARD_DISPLAY

# Import transcendental logic functions
from picounits.core.quantities.scalars.methods import transcendental as tlops


@dataclass(slots=True, repr=False, unsafe_hash=True)
class RealPacket(ScalarPacket):
    """
    A Real Packet: A prefix, value (integer or float) and Unit

    NOTE: Prefix is init-only, value is held in absolute form
    """
    def __post_init__(self, prefix: PrefixScale) -> None:
        """ Validates value and unit, then mutates value to BASE """
        if not isinstance(self.value, (int, float, integer, floating)):
            factory = import_factory("RealPacket.__post_init__")

            # Attempts to pass the value to the correct type
            return factory.create(self.value, self.unit, prefix)

        if not isinstance(self.unit, Unit):
            msg = f"Unit must be of type 'Unit', not {type(self.unit)}"
            raise TypeError(msg)

        if not isinstance(prefix, PrefixScale):
            msg = f"Prefix must be of type PrefixScale, not {type(prefix)}"
            raise TypeError(msg)

        # Mutates prefix to PrefixScale.BASE and scales value
        if prefix == PrefixScale.BASE:
            return

        # Ex.  Kilo (3) - BASE (0) = 3 Hence scaling of 10^3
        prefix_difference = prefix.value - PrefixScale.BASE.value
        factor = 10 ** prefix_difference
        self.value *= factor

    @property
    def name(self) -> str:
        """ Returns the packet name as value + prefix(unit) """
        value, prefix = self._normalize()
        return f"{round(value, STANDARD_DISPLAY)} {prefix}({self.unit.name})"

    @property
    def magnitude(self) -> int | float:
        """ Returns the mathematical absolute value """
        factory = import_factory("RealPacket.magnitude")
        return factory.create(abs(self.value), self.unit)

    @property
    def sign(self) -> int:
        """ Returns the sign of self.value """
        if self.value > 0:
            return 1
        elif self.value < 0:
            return -1
        else:
            return 0

    def _normalize(self) -> tuple[float | int, PrefixScale]:
        """ Normalizes the value for packet name representation """
        value = self.value

        if value == 0 or self.unit == Unit.dimensionless():
            # Handles division by zero or dimensionless values
            return value, PrefixScale.BASE

        # Extraction of exponent
        magnitude = abs(value)
        prefix_power = int(log10(magnitude))

        # Snaps prefix power to multiple of 3
        prefix_power = 3 * (prefix_power // 3)

        # O(n) prefix lookup & calculation of new value
        closest = PrefixScale.from_value(prefix_power)
        value /= 10 ** closest.value

        return value, closest

    def __format__(self, format_spec: str) -> str:
        """ Formats the string based on user input through 'format_spec'"""
        value, prefix = self._normalize()
        formatted_value = format(value, format_spec)

        return f"{formatted_value} {prefix}({self.unit.name})"

    def __ceil__(self) -> Packet:
        """ Defines the behavior for ceiling method """
        factory = import_factory("RealPacket.__ceil__")
        return factory.create(ceil(self.value), self.unit)

    def __floor__(self) -> Packet:
        """ Defines the behavior for floor method """
        factory = import_factory("RealPacket.__floor__")
        return factory.create(floor(self.value), self.unit)

    def __trunc__(self) -> Packet:
        """ Defines the behavior for trunc method """
        factory = import_factory("RealPacket.__trunc__")
        return factory.create(trunc(self.value), self.unit)

    def __round__(self, ndigits=None):
        """ Defines the behavior for the round operation """
        factory = import_factory("RealPacket.__round__")
        return factory.create(round(self.value, ndigits), self.unit)

    def __int__(self) -> int:
        """ Returns the integer representation of the packet value """
        return int(self.value)

    def __eq__(self, other: Any) -> bool:
        """ Defines the behavior for equality comparison """
        q2 = self._get_other_packet(other)
        if self.unit != q2.unit:
            # Unit equality matters
            return False

        return self.value == q2.value

    @staticmethod
    def _valid_comparison(q1: Packet, q2: Packet) -> None:
        """
        Raises a ValueError if q1.unit != q2.unit, if not returns none
        """
        if q1.unit == q2.unit:
            return

        msg = f"Cannot compare different units, {q1.unit} != {q2.unit}"
        raise ValueError(msg)

    def __lt__(self, other: Any) -> bool:
        """ Defines the behavior for less than comparison """
        q2 = self._get_other_packet(other)
        self._valid_comparison(self, q2)

        return self.value < q2.value

    def __le__(self, other: Any) -> bool:
        """ Defines the behavior for less than or equal to comparison """
        q2 = self._get_other_packet(other)
        self._valid_comparison(self, q2)

        return self.value <= q2.value

    def __gt__(self, other: Any) -> bool:
        """ Defines the behavior for greater than comparison """
        q2 = self._get_other_packet(other)
        self._valid_comparison(self, q2)

        return self.value > q2.value

    def __ge__(self, other: Any) -> bool:
        """ Defines the behavior for greater than or equal to comparison """
        q2 = self._get_other_packet(other)
        self._valid_comparison(self, q2)

        return self.value >= q2.value

    def __repr__(self) -> str:
        """ Displays the packet name """
        return str(self.name)

    """
    ================ TRANSCENDENTAL METHODS ================
    """

    def to_radians(self) -> Packet:
        """ If dimensionless, converts the Packet to radians """
        return tlops.to_radians_logic(self)

    def to_degrees(self) -> Packet:
        """ If dimensionless, converts the Packet to degrees """
        return tlops.to_degrees_logic(self)

    def sin(self) -> Packet:
        """ If dimensionless, performs the sine operation on self """
        return tlops.sin_logic(self)

    def cos(self) -> Packet:
        """ If dimensionless, performs the cosine operation on self """
        return tlops.cos_logic(self)

    def tan(self) -> Packet:
        """ If dimensionless, performs the tangent operation on self """
        return tlops.tan_logic(self)

    def csc(self) -> Packet:
        """ If dimensionless, performs the cosecant operation on self """
        return tlops.csc_logic(self)

    def sec(self) -> Packet:
        """ If dimensionless, performs the secant operation on self """
        return tlops.sec_logic(self)

    def cot(self) -> Packet:
        """ If dimensionless, performs the cotangent operation on self """
        return tlops.cot_logic(self)

    def asin(self) -> Packet:
        """ If dimensionless, performs the arc sine operation on self """
        return tlops.asin_logic(self)

    def acos(self) -> Packet:
        """ If dimensionless, performs the arc cosine operation on self """
        return tlops.acos_logic(self)

    def atan(self) -> Packet:
        """ If dimensionless, performs the arc tangent operation on self """
        return tlops.atan_logic(self)

    def atan2(self, other: Any) -> Packet:
        """ Defines the atan2 method for packets """
        q2 = self._get_other_packet(other)
        return tlops.atan2_logic(self, q2)

    def acsc(self) -> Packet:
        """ If dimensionless, performs the arc cosecant operation on self """
        return tlops.acsc_logic(self)

    def asec(self) -> Packet:
        """ If dimensionless, performs the arc secant operation on self """
        return tlops.asec_logic(self)

    def acot(self) -> Packet:
        """ If dimensionless, performs the arc cotangent operation on self """
        return tlops.acot_logic(self)

    def sinh(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic sine operation on self
        """
        return tlops.sinh_logic(self)

    def cosh(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic cosine operation on self
        """
        return tlops.cosh_logic(self)

    def tanh(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic tangent operation on self
        """
        return tlops.tanh_logic(self)

    def csch(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic cosecant operation on self
        """
        return tlops.csch_logic(self)

    def sech(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic secant operation on self
        """
        return tlops.sech_logic(self)

    def coth(self) -> Packet:
        """
        If dimensionless, performs the hyperbolic cotangent operation on self
        """
        return tlops.coth_logic(self)

    def asinh(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic sine operation on
        self
        """
        return tlops.asinh_logic(self)

    def acosh(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic cosine operation on
        self
        """
        return tlops.acosh_logic(self)

    def atanh(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic tangent operation on
        self
        """
        return tlops.atanh_logic(self)

    def acsch(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic cosecant operation on
        self
        """
        return tlops.acsch_logic(self)

    def asech(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic secant operation on
        self
        """
        return tlops.asech_logic(self)

    def acoth(self) -> Packet:
        """
        If dimensionless, performs the inverse hyperbolic cotangent operation
        on self
        """
        return tlops.acoth_logic(self)

    def exp(self) -> Packet:
        """ If dimensionless, performs the exponential operation on self """
        return tlops.exp_logic(self)

    def log(self, base: float | int) -> Packet:
        """ If dimensionless, performs the variable logarithm on self """
        return tlops.log_logic(self, base)

    def log2(self) -> Packet:
        """ If dimensionless, performs the base 2 logarithm on self """
        return tlops.log2_logic(self)

    def log10(self) -> Packet:
        """ If dimensionless, performs the base 10 logarithm on self """
        return tlops.log10_logic(self)

    def nlog(self) -> Packet:
        """ If dimensionless, performs the natural logarithm on self """
        return tlops.nlog_logic(self)



"""
Filename: array.py
Author: William Bowley
Version: 0.1
Clear: X

Description:
    Defines the Array Packet Class which is
    comprised of a Value, Unit and prefixScale.
"""


from __future__ import annotations
from dataclasses import dataclass

from numpy import (
    array, ndarray, floating, integer, ceil, linalg, floor,
    log10, isfinite, abs as np_abs, max as np_max, append as np_append,
    round as np_round
)

from picounits.core.unit import Unit
from picounits.core.scales import PrefixScale

from picounits.core.quantities.packet import Packet
from picounits.core.quantities.vectors.vector import VectorPacket

from picounits.lazy_imports import import_factory
from picounits.configuration.picounits import STANDARD_DISPLAY

@dataclass(slots=True, repr=False, unsafe_hash=True)
class ArrayPacket(VectorPacket):
    """
    A Array Packet: A prefix, array and a unit

    NOTE: Prefix is init-only, value is held in absolute form
    """
    def __post_init__(self, prefix: PrefixScale) -> None:
        """ Validates value and unit, then mutates values to BASE """
        if not isinstance(self.value, (list, tuple, ndarray)):
            factory = import_factory("ArrayPacket.__post_init__")

            # Attempts to pass the value to the correct type
            return factory.create(self.value, self.unit, prefix)

        if not isinstance(self.unit, Unit):
            msg = f"Unit must be of type 'Unit', not {type(self.unit)}"
            raise TypeError(msg)

        if not isinstance(prefix, PrefixScale):
            msg = f"Prefix must be of type PrefixScale, not {type(prefix)}"
            raise TypeError(msg)

        # Mutates prefix to PrefixScale.BASE and scales value
        factor = 1
        if prefix != PrefixScale.BASE:
            # Ex.  Kilo (3) - BASE (0) = 3 Hence scaling of 10^3
            prefix_difference = prefix.value - PrefixScale.BASE.value
            factor = 10 ** prefix_difference

        # Converts the input to a non-scaled ndarry
        self._quantity_conversion(factor)

    def _quantity_conversion(self, factor: int) -> None:
        """Strip packet wrappers and ensure we have numeric types."""
        new_value: list = []
        for item in self.value:
            if isinstance(item, Packet):
                self.unit_check(item) # Use Packet's built-in unit checker
                new_value.append(item.value * factor)

            elif isinstance(item, (int, float, integer, floating)):
                new_value.append(item)

            else:
                if isinstance(item, complex):
                    msg = "Cannot create a vector of complex numbers"
                    raise TypeError(msg)

                msg = f"Cannot convert {type(item)} to ArrayPacket value"
                raise TypeError(msg)

        self.value = array(new_value) * factor

    @property
    def name(self) -> str:
        """ Returns the packet name as value + prefix(unit) """
        value, prefix = self._normalize()
        return f"{np_round(value, STANDARD_DISPLAY)} {prefix}({self.unit.name})"

    @property
    def magnitude(self) -> float:
        """ Returns the magnitude of vector with units """
        factory = import_factory("ArrayPacket.magnitude")

        value = float(linalg.norm(self.value))
        return factory.create(value, self.unit)

    def append(self, item: Packet) -> None:
        """ Appends a Packet to the array. """
        if not isinstance(item, Packet):
            raise TypeError(f"append() argument must be a Packet, not {type(item).__name__}")

        if isinstance(item.value, complex):
            raise TypeError(f"Vectors cannot element thats a complex number, got {item}")

        self.unit_check(item)
        self.value = np_append(self.value, item.value)

    def _normalize(self) -> tuple[complex, PrefixScale]:
        """ Normalizes the value for packet name representation """
        if self.value.size == 0 or self.unit == Unit.dimensionless():
            return self.value, PrefixScale.BASE

        # Focus on the 'peak' magnitude to define the scale
        max_mag = np_max(np_abs(self.value))

        if max_mag == 0 or not isfinite(max_mag):
            return self.value, PrefixScale.BASE

        # Get the prefix for the largest element
        peak_power = int(floor(log10(max_mag)))

        # Snaps prefix power to multiple of 3 & performs O(n) prefix lookup
        prefix_power = 3 * (peak_power // 3)
        closest = PrefixScale.from_value(prefix_power)

        return self.value / (10 ** closest.value), closest

    def __format__(self, format_spec: str) -> str:
        """ Formats the string based on user input through 'format_spec'"""
        value, prefix = self._normalize()
        formatted_value = format(value, format_spec)

        return f"{formatted_value} {prefix}({self.unit.name})"

    def __ceil__(self) -> Packet:
        """ Defines the behavior for ceiling method """
        factory = import_factory("ArrayPacket.__ceil__")
        return factory.create(ceil(self.value), self.unit)

    def __repr__(self) -> str:
        """ Displays the packet name """
        return str(self.name)



"""
Filename: main.py
Author: William Bowley

Description:
    Example of a thermistor linearizer based on 
    SBOA323A by Texas Instruments.

    Calculates the resistors needed to produce
    a linear output voltage over a temperature
    range.
"""

from math import exp

from picounits.core import unit_validator, Quantity as q
from picounits import IMPEDANCE, VOLTAGE, TEMPERATURE, KILO, MILLI

# Thermistor Steinhart-Hart Inverse Coefficients
A1 = -14.6337
B1 = 4791.84
C1 = -115334
D1 = -3.7305e+06

# Circuit Variables
SUPPLY = 3.3 * VOLTAGE
V_OUT_MAX = 3.2 * VOLTAGE
V_OUT_MIN = 100 * MILLI * VOLTAGE

MAX_TEMPERATURE = 125 * TEMPERATURE
MIN_TEMPERATURE = 25 * TEMPERATURE

NTC100AT25C = 100 * KILO * IMPEDANCE

# Standard value for the feedback resistor
FEEDBACK = 1.5 * KILO * IMPEDANCE


@unit_validator(IMPEDANCE)
def calculate_ntc_resistance(temp_celsius: q) -> q:
    """ Calculates the NTC thermistor resistance using Steinhart-Hart """
    tk = temp_celsius.value + 273.15
    exponent_value = (A1 + (B1 / tk) + (C1 / tk ** 2) + (D1 / tk ** 3))

    return NTC100AT25C * exp(exponent_value)


@unit_validator(VOLTAGE)
def voltage_range(low_side_ntc) -> tuple[q]:
    """ Calculates the input voltage range, minium and maximum """
    def calculation(temp_res):
        """ Calculation for voltage depends on different temp res"""
        return SUPPLY * (low_side_ntc / (temp_res + low_side_ntc))

    return calculation(R_NTC_MIN), calculation(R_NTC_MAX)


@unit_validator(IMPEDANCE)
def ref_divider_resistance(
    v_in_max: q, ideal_gain: q, divider: q
) -> tuple[q]:
    """ Calculates the resistance of the high_side and low_side resistors """
    high_side = FEEDBACK * SUPPLY / (v_in_max * ideal_gain - V_OUT_MAX)
    low_side = divider * high_side / (high_side - divider)
    return high_side, low_side


# Calculated values for resistance at max temp and min temp
R_NTC_MAX = calculate_ntc_resistance(MAX_TEMPERATURE)
R_NTC_MIN = calculate_ntc_resistance(MIN_TEMPERATURE)

low_side = (R_NTC_MIN * R_NTC_MAX) ** 0.5
v_in_min, v_in_max = voltage_range(low_side)

ideal_gain = (V_OUT_MAX - V_OUT_MIN) / (v_in_max - v_in_min)
ref_divider = FEEDBACK / (ideal_gain - 1)
ref_high, ref_low = ref_divider_resistance(v_in_max, ideal_gain, ref_divider)
actual_gain = (ref_divider + FEEDBACK) / ref_divider


print("============================")
print(f"Resistance at {MIN_TEMPERATURE.value} c is {R_NTC_MIN:.3f}")
print(f"Resistance at {MAX_TEMPERATURE.value} c is {R_NTC_MAX:.3f}")
print("============================")

# Print out names based partly on naming in the schematic present in (SBOA323A)
print(f"Low side NTC divider (R1) = {low_side:.3f}")
print(f"Low side reference divider (R2) = {ref_low:.3f}")
print(f"High side reference divider (R3) = {ref_high:.3f}")
print("============================")
print(f"Op Amp feedback (R4) = {FEEDBACK:.3f}")
print(f"Op Amp gain = {actual_gain:.3f}")
print("============================")



"""
Filename: vectors.py
Author: William Bowley
Version: 0.1
Clear: X

Description:
    Defines the semi-abstract VectorPacket Class
    which is defines arithmetic routing for vector
    but not representation or prefix scaling
"""

from abc import ABC
from typing import Any, Generator
from dataclasses import dataclass

from picounits.core.unit import Unit
from picounits.core.quantities.packet import Packet
from picounits.core.quantities.factory import Factory

from picounits.core.quantities.vectors.methods import arithmetic as acops
from picounits.core.quantities.vectors.methods import operations as osops


@dataclass(slots=True)
class VectorPacket(Packet, ABC):
    """
    An Abstract Vector Packet: A prefix, value (Any Vector) and Unit

    NOTE: Representation, prefix scaling, comparison, validation
    are not implemented in this base case.
    """

    @property
    def unit_vector(self) -> Packet:
        """ Calculates the unit vector of self """
        return osops.normalize(self)

    def dot(self, other: Packet) -> Packet:
        """ Defines the behavior for the dot product method """
        return osops.dot(self, other)

    def cross(self, other: Packet) -> Packet:
        """ Defines the behavior for the cross product method"""
        return osops.cross(self, other)

    def angle_between(self, other: Packet) -> Packet:
        """
        Defines the behavior for the angle between method. Returns in radians
        """
        return osops.angle_between(self, other)

    def __add__(self, other: Any) -> Packet:
        """ Defines the behavior for the forwards addition operator (+) """
        q2 = self._get_other_packet(other)
        return acops.add_logic(self, q2)

    def __radd__(self, other: Any) -> Packet:
        """ Defines the behavior for the reverse addition operator (+) """
        # Due to the commutative property of addition (a+b = b+a)
        return self.__add__(other)

    def __iadd__(self, other: Any) -> Packet:
        """ Defines in-place addition operation (+=) """
        # Due to the commutative property of addition (a+b = b+a)
        return self.__add__(other)

    def __sub__(self, other: Any) -> Packet:
        """ Defines behavior for the forwards subtraction operator (-) """
        q2 = self._get_other_packet(other)
        return acops.sub_logic(self, q2)

    def __rsub__(self, other: Any) -> Packet:
        """ Defines the behavior for the reverse subtraction method """
        # Due to subtraction being non-commutative
        q1 = self._get_other_packet(other)
        return q1.__sub__(self)

    def __isub__(self, other: Any) -> Packet:
        """ Defines in-place subtraction operation (-=) """
        return self.__sub__(other)

    def __mul__(self, other: Any) -> Packet:
        """
        Defines behavior for the forward multiplication (*)
        Also defines the syntactic bridge to move units into quantity:
        Ex. (1+1j) (s) * m = (1+1j) (s) * 1 (m) => (1+1j) (ms)
        """
        if isinstance(other, Unit):
            q2 = Factory.create(1, other)
        else:
            q2 = self._get_other_packet(other)

        return acops.multiplication_logic(self, q2)

    def __rmul__(self, other: Any) -> Packet:
        """ Defines behavior for the reverse multiplication """
        # Due to the commutative property of multiplication (ab = ba)
        return self.__mul__(other)

    def __imul__(self, other: Any) -> Packet:
        """ Defines in-place multiplication operation (*=) """
        return self.__mul__(other)

    def __truediv__(self, other: Any) -> Packet:
        """
        Defines behavior for the forward true division (/)
        Also defines the syntactic bridge to move units into quantity
        Ex 10 * m => 10 (m) / s = 10 (ms⁻¹)
        """
        if isinstance(other, Unit):
            q2 = Factory.create(1, other)
        else:
            q2 = self._get_other_packet(other)
        return acops.true_division_logic(self, q2)

    def __rtruediv__(self, other: float | int) -> Packet:
        """ Defines behavior for the reverse true division """
        # Due to division being non-commutative
        q1 = self._get_other_packet(other)
        return q1.__truediv__(self)

    def __itruediv__(self, other: Any) -> Packet:
        """ Defines in-place division (/=) """
        return self.__truediv__(other)

    def __pow__(self, other: Any) -> Packet:
        """ Defines behavior for the forward power operator (**) """
        q2 = self._get_other_packet(other)
        return acops.power_logic(self, q2)

    def __rpow__(self, other: float | int) -> Packet:
        """ Defines behavior for the reverse power """
        q1 = self._get_other_packet(other)
        return q1.__pow__(self)

    def _raise_ordering_error(self) -> None:
        """ Helper for ordering comparison errors """
        msg = (
            "Cannot order vector quantities. "
            "Vectors have no natural ordering. "
            "Use abs() or .magnitude to compare magnitudes."
        )
        raise TypeError(msg)

    def __lt__(self, other: Any) -> bool:
        """ Defines the behavior for less than comparison """
        _ = other
        self._raise_ordering_error()

    def __le__(self, other: Any) -> bool:
        """ Defines the behavior for less than or equal to comparison """
        _ = other
        self._raise_ordering_error()

    def __gt__(self, other: Any) -> bool:
        """ Defines the behavior for greater than comparison """
        _ = other
        self._raise_ordering_error()

    def __ge__(self, other: Any) -> bool:
        """ Defines the behavior for greater than or equal to comparison """
        _ = other
        self._raise_ordering_error()

    def __abs__(self) -> Packet:
        """ Defines the absolute value operator """
        return Factory.create(self.magnitude, self.unit)

    def __neg__(self) -> Packet:
        """ Defines behavior for negation operator (-quantity) """
        return Factory.create(-self.value, self.unit)

    def __pos__(self) -> Packet:
        """ Defines behavior for unary plus operator (+quantity) """
        return Factory.create(+self.value, self.unit)

    def __bool__(self) -> bool:
        """ Defines behavior for boolean conversion (USES MAGNITUDE) """
        return self.magnitude != 0

    def __len__(self) -> int:
        """ Returns the number of elements in the vector. """
        return len(self.value)

    def __getitem__(self, index) -> Any:
        """ Allows indexing like packet[0] """
        return Factory.create(self.value[index], self.unit)

    def __setitem__(self, index: int | slice, other: Any) -> None:
        """ 
        Allows item assignment like packet[0] = value 
        """
        if isinstance(other, Packet):
            if not isinstance(other.value, complex):
                self.unit_check(other)
                value_to_set = other.value

                self.value[index] = value_to_set
                return

        msg = f"{type(other)!r} is an unsupported type for item assignment"
        raise TypeError(msg)

    def __iter__(self) -> Generator[Packet, None, None]:
        """ Returns an iterator for the array. """
        for val in self.value:
            yield Factory.create(val, self.unit)



"""
Filename: priority.py
Author: William Bowley
Version: 0.1

Description:
    Builds the domain priority dictionary for
    factory.

    NOTE: Due to circular imports this dictionary
    has to be in a separate file and loaded in
    during Factory.reallocate() via lazy imports.
"""

from picounits.core.quantities.scalars.scalar import ScalarPacket
from picounits.core.quantities.vectors.vector import VectorPacket


DOMAIN_PRIORITY = {
    ScalarPacket: 1,
    VectorPacket: 2,
    # Add new ones here - just assign the Arithmetic Router
}



"""
Filename: engine.py
Author: William Bowley

Description:
    Transient thermal solver for the 2D heat equation PDE
    using picounits to show unit-boundary abstractions.
    
    NOTE:
    This is a demonstration model and is not intended to 
    match a refined finite element solver
"""

import numpy as np

from shapely import box, MultiPoint
from shapely.ops import triangulate

from picounits import UnitError, Q, strip_quantity
from picounits.extensions.loader import DynamicLoader

from picounits.constants import (
    LENGTH, TEMPERATURE, TIME, VOLUMETRIC_HEATING,
    VOLUMETRIC_HEAT_CAPACITY, THERMAL_CONDUCTIVITY,
    CONVECTION_COEFFICIENT, DIMENSIONLESS
)


class PreProcessor:
    """ Builds geometry and then meshes the problem domain """
    @staticmethod
    def rectangle(bottom_left: tuple[Q, Q], length: Q, height: Q) -> list:
        """ Creates a square vector geometry """
        x, y = bottom_left

        for item in (x, y, length, height):
            if isinstance(item, Q):
                if item.unit == LENGTH:
                    break

            msg = f"{item!r} must be have unit {LENGTH} not {item.unit!r}"
            raise UnitError(msg)

        return [x.value, y.value, (x + length).value, (y + height).value]

    def __init__(self, domain: list, source: list, density: int = 30) -> None:
        """ Constructs mesh based on geometry built via static methods """
        # Saves raw coordinates
        self.domain_coordinates = domain

        # Converts source to shapely box
        source = box(source[0], source[1], source[2], source[3])
        self.source_box_coords = list(source.exterior.coords)

        # Builds grid around geometry
        bx, by, tx, ty = domain
        x_coords, y_coords = np.linspace(bx, tx, density), np.linspace(by, ty, density)
        points = MultiPoint([(x, y) for x in x_coords for y in y_coords])

        # Meshes the geometry
        all_points_list = [(p.x, p.y) for p in points.geoms] + self.source_box_coords
        refined_points = MultiPoint(all_points_list)

        # Saves mesh
        self.mesh = triangulate(refined_points)
        self.source_box = source


class Solver:
    """ Solves the transient thermal PDE over 2D geometry using FEM """
    def __init__(self, preprocessor: PreProcessor) -> None:
        """ Initializes the solver """
        # Saves mesh and raw geometry for matrix assembly
        self.mesh = preprocessor.mesh
        self.domain = preprocessor.domain_coordinates
        self.source = preprocessor.source_box

        # Calculates the number of nodes and elements
        nodes = []
        for triangle in self.mesh:
            nodes.extend(list(triangle.exterior.coords))

        self.unique_nodes = np.unique(nodes, axis=0)
        self.num_nodes = len(self.unique_nodes)

        self.num_element = len(self.mesh)

        # Configures the conductivity array, load vector, mass and stiffness matrix
        self.connectivity = []
        self.load_vector = np.zeros(self.num_nodes)
        self.m_stiffness = np.zeros((self.num_nodes, self.num_nodes))
        self.m_mass = np.zeros((self.num_nodes, self.num_nodes))

        # Post init variables
        self.constructed = False
        self.source_capacity = None
        self.convection = None
        self.source_dependence = None

        self.domain_capacity = None
        self.domain_dependence = None

        self.time_step = 0
        self.axial_length = 0
        self.boundary_temp = 0
        self.heating = 0
        self.tolerance = 0

        self.system = None
        self.connectivity = None
        self.boundary_edges = None
        self.interface_edges = None
        self.boundary_indices: np.ndarray = None

    def build(self, params: DynamicLoader) -> None:
        """ Builds the simulation domain via assembly of stiffness and mass matrixes """
        # Extracts materials properties
        def _decompile(material: DynamicLoader) -> tuple[list, float]:
            """ Validates and decompile's dependence table """
            temp, k_vals = [], []
            for group in material.temp_dependence:
                temperature = strip_quantity(group[0], TEMPERATURE)
                conductivity = strip_quantity(group[1], THERMAL_CONDUCTIVITY)

                temp.append(temperature)
                k_vals.append(conductivity)

            capacity = material.volumetric_heat_capacity
            capacity = strip_quantity(capacity, VOLUMETRIC_HEAT_CAPACITY)
            return (temp, k_vals), capacity

        # Decompile's materials
        self.source_dependence, self.source_capacity = _decompile(params.copper)
        self.domain_dependence, self.domain_capacity = _decompile(params.air)
        convection = params.problem.convection

        # Extracts and strips away units after validation
        self.time_step = strip_quantity(params.problem.time_step, TIME)
        self.axial_length = strip_quantity(params.problem.axial_length, LENGTH)
        self.convection = strip_quantity(convection, CONVECTION_COEFFICIENT)
        self.boundary_temp = strip_quantity(params.problem.temperature, TEMPERATURE)
        self.heating = strip_quantity(params.source.power_density, VOLUMETRIC_HEATING)
        self.tolerance = strip_quantity(params.problem.tolerance, DIMENSIONLESS)

        self._matrix_assembly()

    def solve(
        self,
        current_solution: np.ndarray | None = None,
        max_iterations: int = 1000,
    ) -> np.ndarray:
        """" Solves the thermal problem for the current time step """
        # Configures the boundary conditions
        x_min, y_min, x_max, y_max = self.domain

        # Creates the boundary indices using tolerance on first iteration
        if self.boundary_indices is None:
            unique_nodes = self.unique_nodes
            self.boundary_indices = np.where(
                (np.abs(unique_nodes[:, 0] - x_min) < self.tolerance) |
                (np.abs(unique_nodes[:, 0] - x_max) < self.tolerance) |
                (np.abs(unique_nodes[:, 1] - y_min) < self.tolerance) |
                (np.abs(unique_nodes[:, 1] - y_max) < self.tolerance)
            )[0]

        if current_solution is None:
            u_old = np.full(self.num_nodes, self.boundary_temp)
        else:
            u_old = current_solution

        u_crr = u_old.copy()
        for _ in range(max_iterations):
            self.m_stiffness.fill(0)
            self.load_vector.fill(0)
            self._matrix_assembly(u_crr)

            # Form the backwards euler equation
            self.system = self.m_mass + self.time_step * self.m_stiffness
            b_rhs = self.m_mass @ u_old + self.time_step * self.load_vector

            u_next = self._gradient_solver(self.system, b_rhs, u_crr, self.tolerance)

            if np.linalg.norm(u_next - u_crr) < self.tolerance:
                return u_next

        return u_next

    def _matrix_assembly(self, current_solution=None):
        """ Constructs the initial matrix and updates elements per time step"""
        if not self.constructed:
            # Assembles the mesh for first solve
            self._preprocess_mesh()

        # Adds the heat source per time step
        for index, triangle in enumerate(self.mesh):
            indices = self.connectivity[index] if self.constructed else None
            self._assemble_element(index, triangle, indices, current_solution)

        # Applies convection to mass matrix per time step
        self._apply_convection()

        if not self.constructed:
            self.constructed = True

    def _preprocess_mesh(self):
        """ Builds the connectivity map and edge lists based on mesh"""
        x_min, y_min, x_max, y_max = self.domain
        s_xmin, s_ymin, s_xmax, s_ymax = self.source.bounds
        tol = self.tolerance

        conductivity_map = []
        self.boundary_edges = []
        self.interface_edges = []
        for triangle in self.mesh:
            coords = np.array(triangle.exterior.coords)[:3]
            indices = [
                np.argmin(np.linalg.norm(self.unique_nodes - p, axis=1)) for p in coords
            ]

            # Checks if coords of the triangle is on the boundary of source or domain
            for i, j in [(0, 1), (1, 2), (2, 0)]:
                p1, p2 = coords[i], coords[j]
                edge_len = np.linalg.norm(p1 - p2)

                on_boundary = (
                    self._on_line(p1, p2, 'x', x_min, y_min, y_max, tol) or
                    self._on_line(p1, p2, 'x', x_max, y_min, y_max, tol) or
                    self._on_line(p1, p2, 'y', y_min, x_min, x_max, tol) or
                    self._on_line(p1, p2, 'y', y_max, x_min, x_max, tol)
                )
                if on_boundary:
                    self.boundary_edges.append(([indices[i], indices[j]], edge_len))

                on_source = (
                    self._on_line(p1, p2, 'x', s_xmin, s_ymin, s_ymax, tol) or
                    self._on_line(p1, p2, 'x', s_xmax, s_ymin, s_ymax, tol) or
                    self._on_line(p1, p2, 'y', s_ymin, s_xmin, s_xmax, tol) or
                    self._on_line(p1, p2, 'y', s_ymax, s_xmin, s_xmax, tol)
                )
                if on_source:
                    self.interface_edges.append(([indices[i], indices[j]], edge_len))

            conductivity_map.append(indices)
        self.connectivity = np.array(conductivity_map)

    def _assemble_element(self, index, triangle, indices, current_solution):
        """ Assembles stiffness, mass, and load contributions for a single element """
        coords = np.array(triangle.exterior.coords)[:3]
        indices = self.connectivity[index]

        b, c, area = self._calculate_triangle(coords)
        grid = np.ix_(indices, indices)
        is_source = triangle.intersects(self.source)

        if is_source:
            self._add_heating(indices, area)
            self._add_face_cooling(indices, grid, area)

        self._add_conduction(indices, grid, area, b, c, current_solution, is_source)

        if not self.constructed:
            self._add_mass(grid, area, is_source)

    def _add_heating(self, indices, area):
        """ Adds volumetric heat generation to load vector """
        heat_per_node = self.heating * self.axial_length * area / 3
        for i in indices:
            self.load_vector[i] += heat_per_node

    def _add_face_cooling(self, indices, grid, area):
        """ Adds out-of-plane convective cooling on source faces """
        h = self.convection
        ke_face = (h * area / 12) * np.array([[2, 1, 1], [1, 2, 1], [1, 1, 2]])
        fe_face = (h * self.boundary_temp * area / 3) * np.array([1, 1, 1])
        self.m_stiffness[grid] += ke_face
        self.load_vector[indices] += fe_face

    def _add_conduction(self, indices, grid, area, b, c, current_solution, is_source):
        """ Adds temperature-dependent conduction to stiffness matrix """
        temp_dep = self.source_dependence if is_source else self.domain_dependence
        u_elem = (
            current_solution[indices]
            if current_solution is not None else self.boundary_temp
        )

        k_temp = np.interp(np.mean(u_elem), temp_dep[0], temp_dep[1])
        local_b = np.array([b, c]) / (2 * area)
        ke_cond = k_temp * self.axial_length * area * np.dot(local_b.T, local_b)
        self.m_stiffness[grid] += ke_cond

    def _add_mass(self, grid, area, is_source):
        """ Adds thermal inertia to mass matrix (first assembly only) """
        h_cap = self.source_capacity if is_source else self.domain_capacity

        me_map = np.array([[2, 1, 1], [1, 2, 1], [1, 1, 2]])
        me_local = h_cap * self.axial_length * (area / 12) * me_map
        self.m_mass[grid] += me_local

    def _apply_convection(self):
        """ Applies convective boundary conditions to stiffness matrix and load vector """
        h, T_inf, axial = self.convection, self.boundary_temp, self.axial_length

        for edge_indices, length in (self.boundary_edges + self.interface_edges):
            ke_conv = (h * length * axial / 6) * np.array([[2, 1], [1, 2]])
            fe_conv = (h * T_inf * length * axial / 2) * np.array([1, 1])
            grid = np.ix_(edge_indices, edge_indices)
            self.m_stiffness[grid] += ke_conv
            self.load_vector[edge_indices] += fe_conv

    def _on_line(self, p1, p2, axis, value, lo, hi, tol):
        """ Checks if both points lie on an axis-aligned edge within bounds """
        a, b = (0, 1) if axis == 'x' else (1, 0)
        return (
            np.abs(p1[a] - value) < tol and 
            np.abs(p2[a] - value) < tol and
            lo - tol <= p1[b] <= hi + tol and
            lo - tol <= p2[b] <= hi + tol
        )

    def _calculate_triangle(self, coords: np.ndarray) -> tuple[list, list, float]:
        """ Calculates piecewise linear basis and triangle area via Shoelace Formula """
        # Breaks nodes into coordinates
        x1, y1 = coords[0]
        x2, y2 = coords[1]
        x3, y3 = coords[2]

        # Uses coordinates to calculate (b,c) matrix and area
        b = [y2-y3, y3-y1, y1-y2]
        c = [x3-x2, x1-x3, x2-x1]
        area = 0.5 * abs(x1*b[0] + x2*b[1] + x3*b[2])

        return b, c, area

    def _gradient_solver(
        self,
        A: np.ndarray,
        B: np.ndarray,
        sol: np.ndarray | None = None,
        tolerance: float = 1e-9,
        max_iter: int =1000
    ) -> np.ndarray:
        """ Solvers Ax=b via conjugate gradient solver for Ax=B linear system """
        # Uses old solution as initial guess or empty B matrix
        sol = sol if sol is not None else np.zeros_like(B)

        # Distance A.sol is from B
        distance = B - np.dot(A, sol)
        direction = distance.copy()
        magnitude = np.dot(distance, distance)

        for _ in range(max_iter):
            # Calculates the step size (alpha)
            A_d = np.dot(A, direction)
            alpha = magnitude / np.dot(direction, A_d)

            # Update solution via moving solution by alpha along direction
            sol += alpha * direction
            distance -= alpha * A_d

            # Check magnitude is less than tolerance
            new_magnitude = np.dot(distance, distance)
            if np.sqrt(new_magnitude) < tolerance:
                break

            # Updates search direction and magnitude
            direction = distance + (new_magnitude / magnitude) * direction
            magnitude = new_magnitude

        return sol



"""
Filename: factory.py
Author: William Bowley
Version: 0.1
Clear: X

Description:
    Defines the Quantity Factory Class, acts
    as a type caster for quantities within
    the agnostic functional endpoint methods.
"""

from typing import Any, Callable
from numpy import ndarray, integer, floating, complexfloating
from picounits.core.quantities.packet import Packet

from picounits.lazy_imports import lazy_import

class Factory:
    """
    Packet type casing factory for usage in agnostic functional methods
    """
    @classmethod
    def create(cls, value: Any, unit, prefix=None) -> Packet:
        """
        Finds the type of the value and returns a casted packet
        NOTE: The usage of import injection is to avoid circular imports
        NOTE: Cannot type hint unit nor prefix due to circular imports
        """
        if prefix is None:
            prefixscale = lazy_import(
                "picounits.core.scales", "PrefixScale", "Factory.create"
            )
            prefix = prefixscale.BASE

        match value:
            case complex() | complexfloating():
                complex_packet = lazy_import(
                    "picounits.core.quantities.scalars.types.complex", 
                    "ComplexPacket", "Factory.create"
                )

                return complex_packet(value, unit, prefix)

            case float() | int() | integer() | floating():
                real_packet = lazy_import(
                    "picounits.core.quantities.scalars.types.real", 
                    "RealPacket", "Factory.create"
                )
                return real_packet(value, unit, prefix)

            case tuple() | list() | ndarray():
                array_packet = lazy_import(
                    "picounits.core.quantities.vectors.types.array",
                    "ArrayPacket", "Factory.create"
                )
                return array_packet(value, unit, prefix)

            case _:
                msg = f"No Packet for this value type: {type(value)}"
                raise TypeError(msg)

    @classmethod
    def reallocate(cls, op_name: str) -> Callable:
        """
        Reallocate arithmetic or transcendental methods when packets are
        not similar types. For example ScalarPacket != VectorPacket
        """

        def decorator(method: Callable) -> Callable:
            def wrapper(q1: Packet, q2: Packet) -> Callable:
                """ Promotes methods if they dont have the same parents """
                q1_parents = q1.__class__.__bases__[0]
                q2_parents = q2.__class__.__bases__[0]
                if q1_parents == q2_parents:
                    # Simple pass through if same class has the same parent
                    return method(q1, q2)

                # Initialization due to circular import of literals
                domain_priority = lazy_import(
                    "picounits.core.quantities.priority",
                    "DOMAIN_PRIORITY", "Factory.reallocate"
                )

                # Gets priority of each packet_type
                p1 = domain_priority.get(q1_parents, 0)
                p2 = domain_priority.get(q2_parents, 0)

                # Finds the native operator for these packets
                winner = q1 if p1 >= p2 else q2
                loser = q2 if winner is q1 else q1

                if q1.__class__ == winner.__class__:
                    # If winner is q1 than return the method
                    return method(q1, q2)

                # Reallocate this result to another arithmetic router
                return getattr(winner, op_name)(loser)
            return wrapper
        return decorator

    @classmethod
    def category_check(cls, q1: Packet, q2: Packet) -> None:
        """ Checks if two packets are from the same category """
        q1_parents = q1.__class__.__bases__[0]
        q2_parents = q2.__class__.__bases__[0]

        if q1_parents == q2_parents:
            return

        msg = f"{q1!r} is not in the same quality category as {q2!r}"
        raise TypeError(msg)
